import os
import warnings
from functools import partial

import numpy as np

from yt import units
from yt._typing import KnownFieldsT
from yt.fields.field_detector import FieldDetector
from yt.fields.field_info_container import FieldInfoContainer
from yt.frontends.ramses.io import convert_ramses_conformal_time_to_physical_time
from yt.utilities.cython_fortran_utils import FortranFile
from yt.utilities.lib.cosmology_time import t_frw
from yt.utilities.linear_interpolators import BilinearFieldInterpolator
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.physical_constants import (
    boltzmann_constant_cgs,
    mass_hydrogen_cgs,
    mh,
    mp,
)

from .field_handlers import RTFieldFileHandler

b_units = "code_magnetic"
ra_units = "code_length / code_time**2"
rho_units = "code_density"
vel_units = "code_velocity"
pressure_units = "code_pressure"
ener_units = "code_mass * code_velocity**2"
specific_ener_units = "code_velocity**2"
ang_mom_units = "code_mass * code_velocity * code_length"
cooling_function_units = " erg * cm**3 /s"
cooling_function_prime_units = " erg * cm**3 /s/K"
flux_unit = "1 / code_length**2 / code_time"
number_density_unit = "1 / code_length**3"

known_species_masses = {
    sp: mh * v
    for sp, v in [
        ("HI", 1.0),
        ("HII", 1.0),
        ("Electron", 1.0),
        ("HeI", 4.0),
        ("HeII", 4.0),
        ("HeIII", 4.0),
        ("H2I", 2.0),
        ("H2II", 2.0),
        ("HM", 1.0),
        ("DI", 2.0),
        ("DII", 2.0),
        ("HDI", 3.0),
    ]
}

known_species_names = {
    "HI": "H_p0",
    "HII": "H_p1",
    "Electron": "El",
    "HeI": "He_p0",
    "HeII": "He_p1",
    "HeIII": "He_p2",
    "H2I": "H2_p0",
    "H2II": "H2_p1",
    "HM": "H_m1",
    "DI": "D_p0",
    "DII": "D_p1",
    "HDI": "HD_p0",
}

_cool_axes = ("lognH", "logT")  # , "logTeq")
_cool_arrs = (
    ("cooling_primordial", cooling_function_units),
    ("heating_primordial", cooling_function_units),
    ("cooling_compton", cooling_function_units),
    ("heating_compton", cooling_function_units),
    ("cooling_metal", cooling_function_units),
    ("cooling_primordial_prime", cooling_function_prime_units),
    ("heating_primordial_prime", cooling_function_prime_units),
    ("cooling_compton_prime", cooling_function_prime_units),
    ("heating_compton_prime", cooling_function_prime_units),
    ("cooling_metal_prime", cooling_function_prime_units),
    ("mu", None),
    ("abundances", None),
)
_cool_species = (
    "Electron_number_density",
    "HI_number_density",
    "HII_number_density",
    "HeI_number_density",
    "HeII_number_density",
    "HeIII_number_density",
)

_X = 0.76  # H fraction, hardcoded
_Y = 0.24  # He fraction, hardcoded


class RAMSESFieldInfo(FieldInfoContainer):
    known_other_fields: KnownFieldsT = (
        ("Density", (rho_units, ["density"], None)),
        ("x-velocity", (vel_units, ["velocity_x"], None)),
        ("y-velocity", (vel_units, ["velocity_y"], None)),
        ("z-velocity", (vel_units, ["velocity_z"], None)),
        ("Pres_IR", (pressure_units, ["pres_IR", "pressure_IR"], None)),
        ("Pressure", (pressure_units, ["pressure"], None)),
        ("Metallicity", ("", ["metallicity"], None)),
        ("HII", ("", ["H_p1_fraction"], None)),
        ("HeII", ("", ["He_p1_fraction"], None)),
        ("HeIII", ("", ["He_p2_fraction"], None)),
        ("x-acceleration", (ra_units, ["acceleration_x"], None)),
        ("y-acceleration", (ra_units, ["acceleration_y"], None)),
        ("z-acceleration", (ra_units, ["acceleration_z"], None)),
        ("Potential", (specific_ener_units, ["potential"], None)),
        ("B_x_left", (b_units, ["magnetic_field_x_left"], None)),
        ("B_x_right", (b_units, ["magnetic_field_x_right"], None)),
        ("B_y_left", (b_units, ["magnetic_field_y_left"], None)),
        ("B_y_right", (b_units, ["magnetic_field_y_right"], None)),
        ("B_z_left", (b_units, ["magnetic_field_z_left"], None)),
        ("B_z_right", (b_units, ["magnetic_field_z_right"], None)),
    )
    known_particle_fields: KnownFieldsT = (
        ("particle_position_x", ("code_length", [], None)),
        ("particle_position_y", ("code_length", [], None)),
        ("particle_position_z", ("code_length", [], None)),
        ("particle_velocity_x", (vel_units, [], None)),
        ("particle_velocity_y", (vel_units, [], None)),
        ("particle_velocity_z", (vel_units, [], None)),
        ("particle_mass", ("code_mass", [], None)),
        ("particle_identity", ("", ["particle_index"], None)),
        ("particle_refinement_level", ("", [], None)),
        ("particle_birth_time", ("code_time", ["age"], None)),
        ("conformal_birth_time", ("", [], None)),
        ("particle_metallicity", ("", [], None)),
        ("particle_family", ("", [], None)),
        ("particle_tag", ("", [], None)),
        # sink field parameters
        ("particle_mass", ("code_mass", [], None)),
        ("particle_angular_momentum_x", (ang_mom_units, [], None)),
        ("particle_angular_momentum_y", (ang_mom_units, [], None)),
        ("particle_angular_momentum_z", (ang_mom_units, [], None)),
        ("particle_formation_time", ("code_time", [], None)),
        ("particle_accretion_rate", ("code_mass/code_time", [], None)),
        ("particle_delta_mass", ("code_mass", [], None)),
        ("particle_rho_gas", (rho_units, [], None)),
        ("particle_cs**2", (vel_units, [], None)),
        ("particle_etherm", (ener_units, [], None)),
        ("particle_velocity_x_gas", (vel_units, [], None)),
        ("particle_velocity_y_gas", (vel_units, [], None)),
        ("particle_velocity_z_gas", (vel_units, [], None)),
        ("particle_mass_bh", ("code_mass", [], None)),
        ("particle_level", ("", [], None)),
        ("particle_radius_star", ("code_length", [], None)),
    )

    known_sink_fields: KnownFieldsT = (
        ("particle_position_x", ("code_length", [], None)),
        ("particle_position_y", ("code_length", [], None)),
        ("particle_position_z", ("code_length", [], None)),
        ("particle_velocity_x", (vel_units, [], None)),
        ("particle_velocity_y", (vel_units, [], None)),
        ("particle_velocity_z", (vel_units, [], None)),
        ("particle_mass", ("code_mass", [], None)),
        ("particle_identifier", ("", ["particle_index"], None)),
        ("particle_birth_time", ("code_time", ["age"], None)),
        ("BH_real_accretion", ("code_mass/code_time", [], None)),
        ("BH_bondi_accretion", ("code_mass/code_time", [], None)),
        ("BH_eddington_accretion", ("code_mass/code_time", [], None)),
        ("BH_esave", (ener_units, [], None)),
        ("gas_spin_x", (ang_mom_units, [], None)),
        ("gas_spin_y", (ang_mom_units, [], None)),
        ("gas_spin_z", (ang_mom_units, [], None)),
        ("BH_spin_x", ("", [], None)),
        ("BH_spin_y", ("", [], None)),
        ("BH_spin_z", ("", [], None)),
        ("BH_spin", (ang_mom_units, [], None)),
        ("BH_efficiency", ("", [], None)),
    )

    def setup_particle_fields(self, ptype):
        super().setup_particle_fields(ptype)

        def star_age_from_conformal_cosmo(field, data):
            conformal_age = data[ptype, "conformal_birth_time"]
            birth_time = convert_ramses_conformal_time_to_physical_time(
                data.ds, conformal_age
            )
            return data.ds.current_time - birth_time

        def star_age_from_physical_cosmo(field, data):
            H0 = float(
                data.ds.quan(data.ds.hubble_constant * 100, "km/s/Mpc").to("1/Gyr")
            )
            times = data[ptype, "conformal_birth_time"].value
            time_tot = float(t_frw(data.ds, 0) * H0)
            birth_time = (time_tot + times) / H0
            t_out = float(data.ds.current_time.to("Gyr"))
            return data.apply_units(t_out - birth_time, "Gyr")

        def star_age(field, data):
            formation_time = data[ptype, "particle_birth_time"]
            return data.ds.current_time - formation_time

        if self.ds.cosmological_simulation and self.ds.use_conformal_time:
            fun = star_age_from_conformal_cosmo
        elif self.ds.cosmological_simulation:
            fun = star_age_from_physical_cosmo
        else:
            fun = star_age

        self.add_field(
            (ptype, "star_age"),
            sampling_type="particle",
            function=fun,
            units=self.ds.unit_system["time"],
        )

    def setup_fluid_fields(self):
        def _temperature_over_mu(field, data):
            rv = data["gas", "pressure"] / data["gas", "density"]
            rv *= mass_hydrogen_cgs / boltzmann_constant_cgs
            return rv

        self.add_field(
            ("gas", "temperature_over_mu"),
            sampling_type="cell",
            function=_temperature_over_mu,
            units=self.ds.unit_system["temperature"],
        )
        found_cooling_fields = self.create_cooling_fields()

        if found_cooling_fields:

            def _temperature(field, data):
                return data["gas", "temperature_over_mu"] * data["gas", "mu"]

        else:

            def _temperature(field, data):
                if not isinstance(data, FieldDetector):
                    warnings.warn(
                        "Trying to calculate temperature but the cooling tables "
                        "couldn't be found or read. yt will return T/µ instead of "
                        "T — this is equivalent to assuming µ=1.0. To suppress this, "
                        "derive the temperature from temperature_over_mu with "
                        "some values for mu.",
                        category=RuntimeWarning,
                        stacklevel=1,
                    )
                return data["gas", "temperature_over_mu"]

        self.add_field(
            ("gas", "temperature"),
            sampling_type="cell",
            function=_temperature,
            units=self.ds.unit_system["temperature"],
        )

        self.species_names = [
            known_species_names[fn]
            for ft, fn in self.field_list
            if fn in known_species_names
        ]

        # See if we need to load the rt fields
        rt_flag = RTFieldFileHandler.any_exist(self.ds)
        if rt_flag:  # rt run
            self.create_rt_fields()

        # Load magnetic fields
        if ("gas", "magnetic_field_x_left") in self:
            self.create_magnetic_fields()

        # Potential field
        if ("gravity", "Potential") in self:
            self.create_gravity_fields()

    def create_gravity_fields(self):
        def potential_energy(field, data):
            return data["gas", "potential"] * data["gas", "cell_mass"]

        self.add_field(
            ("gas", "potential_energy"),
            sampling_type="cell",
            function=potential_energy,
            units=self.ds.unit_system["energy"],
        )

    def create_magnetic_fields(self):
        # Calculate cell-centred magnetic fields from face-centred
        def mag_field(ax):
            def _mag_field(field, data):
                return (
                    data["gas", f"magnetic_field_{ax}_left"]
                    + data["gas", f"magnetic_field_{ax}_right"]
                ) / 2

            return _mag_field

        for ax in self.ds.coordinates.axis_order:
            self.add_field(
                ("gas", f"magnetic_field_{ax}"),
                sampling_type="cell",
                function=mag_field(ax),
                units=self.ds.unit_system["magnetic_field_cgs"],
            )

        def _divB(field, data):
            """Calculate magnetic field divergence"""
            out = np.zeros_like(data["gas", "magnetic_field_x_right"])
            for ax in data.ds.coordinates.axis_order:
                out += (
                    data["gas", f"magnetic_field_{ax}_right"]
                    - data["gas", f"magnetic_field_{ax}_left"]
                )
            return out / data["gas", "dx"]

        self.add_field(
            ("gas", "magnetic_field_divergence"),
            sampling_type="cell",
            function=_divB,
            units=self.ds.unit_system["magnetic_field_cgs"]
            / self.ds.unit_system["length"],
        )

    def create_rt_fields(self):
        self.ds.fluid_types += ("rt",)
        p = RTFieldFileHandler.get_rt_parameters(self.ds).copy()
        p.update(self.ds.parameters)
        ngroups = p["nGroups"]
        # Make sure rt_c_frac is at least as long as the number of levels in
        # the simulation. Pad with either 1 (default) when using level-dependent
        # reduced speed of light, otherwise pad with a constant value
        if len(p["rt_c_frac"]) == 1:
            pad_value = p["rt_c_frac"][0]
        else:
            pad_value = 1
        rt_c_frac = np.pad(
            p["rt_c_frac"],
            (0, max(0, self.ds.max_level - len(["rt_c_frac"]) + 1)),
            constant_values=pad_value,
        )

        rt_c = rt_c_frac * units.c / (p["unit_l"] / p["unit_t"])
        dens_conv = (p["unit_np"] / rt_c).value / units.cm**3

        ########################################
        # Adding the fields in the hydro_* files
        def _temp_IR(field, data):
            rv = data["gas", "pres_IR"] / data["gas", "density"]
            rv *= mass_hydrogen_cgs / boltzmann_constant_cgs
            return rv

        self.add_field(
            ("gas", "temp_IR"),
            sampling_type="cell",
            function=_temp_IR,
            units=self.ds.unit_system["temperature"],
        )

        def _species_density(field, data, species: str):
            return data["gas", f"{species}_fraction"] * data["gas", "density"]

        def _species_mass(field, data, species: str):
            return data["gas", f"{species}_density"] * data["index", "cell_volume"]

        for species in ["H_p1", "He_p1", "He_p2"]:
            self.add_field(
                ("gas", species + "_density"),
                sampling_type="cell",
                function=partial(_species_density, species=species),
                units=self.ds.unit_system["density"],
            )

            self.add_field(
                ("gas", species + "_mass"),
                sampling_type="cell",
                function=partial(_species_mass, species=species),
                units=self.ds.unit_system["mass"],
            )

        ########################################
        # Adding the fields in the rt_ files
        def gen_pdens(igroup):
            def _photon_density(field, data):
                # The photon density depends on the possibly level-dependent conversion factor.
                ilvl = data["index", "grid_level"].astype("int64")
                dc = dens_conv[ilvl]
                rv = data["ramses-rt", f"Photon_density_{igroup + 1}"] * dc
                return rv

            return _photon_density

        for igroup in range(ngroups):
            self.add_field(
                ("rt", f"photon_density_{igroup + 1}"),
                sampling_type="cell",
                function=gen_pdens(igroup),
                units=self.ds.unit_system["number_density"],
            )

        flux_conv = p["unit_pf"] / units.cm**2 / units.s
        flux_unit = (
            1 / self.ds.unit_system["time"] / self.ds.unit_system["length"] ** 2
        ).units

        def gen_flux(key, igroup):
            def _photon_flux(field, data):
                rv = data["ramses-rt", f"Photon_flux_{key}_{igroup + 1}"] * flux_conv
                return rv

            return _photon_flux

        for key in "xyz":
            for igroup in range(ngroups):
                self.add_field(
                    ("rt", f"photon_flux_{key}_{igroup + 1}"),
                    sampling_type="cell",
                    function=gen_flux(key, igroup),
                    units=flux_unit,
                )

    def create_cooling_fields(self) -> bool:
        "Create cooling fields from the cooling files. Return True if successful."
        num = os.path.basename(self.ds.parameter_filename).split(".")[0].split("_")[1]
        filename = "%s/cooling_%05i.out" % (
            os.path.dirname(self.ds.parameter_filename),
            int(num),
        )

        if not os.path.exists(filename):
            mylog.warning("This output has no cooling fields")
            return False

        # Function to create the cooling fields
        def _create_field(name, interp_object, unit):
            def _func(field, data):
                shape = data["gas", "temperature_over_mu"].shape
                # Ramses assumes a fraction X of Hydrogen within the non-metal gas.
                # It has to be corrected by metallicity.
                Z = data["gas", "metallicity"]
                nH = ((1 - _Y) * (1 - Z) * data["gas", "density"] / mh).to("cm**-3")
                if data.ds.self_shielding:
                    boost = np.maximum(np.exp(-nH / 0.01), 1e-20)
                else:
                    boost = 1
                d = {
                    "lognH": np.log10(nH / boost).ravel(),
                    "logT": np.log10(data["gas", "temperature_over_mu"]).ravel(),
                }
                rv = interp_object(d).reshape(shape)
                if name[-1] != "mu":
                    rv = 10 ** interp_object(d).reshape(shape)
                cool = data.ds.arr(rv, unit)
                if "metal" in name[-1].split("_"):
                    cool = (
                        cool * data["gas", "metallicity"] / 0.02
                    )  # Ramses uses Zsolar=0.02
                elif "compton" in name[-1].split("_"):
                    cool = data.ds.arr(rv, unit + "/cm**3")
                    cool = (
                        cool / data["gas", "number_density"]
                    )  # Compton cooling/heating is written to file in erg/s
                return cool

            self.add_field(name=name, sampling_type="cell", function=_func, units=unit)

        # Load cooling files
        avals = {}
        tvals = {}
        with FortranFile(filename) as fd:
            n1, n2 = fd.read_vector("i")
            for ax in _cool_axes:
                avals[ax] = fd.read_vector("d")
            for i, (tname, unit) in enumerate(_cool_arrs):
                var = fd.read_vector("d")
                if var.size == n1 and i == 0:
                    # If this case occurs, the cooling files were produced pre-2010 in
                    # a format that is no longer supported
                    mylog.warning(
                        "This cooling file format is no longer supported. "
                        "Cooling field loading skipped."
                    )
                    return False
                if var.size == n1 * n2:
                    tvals[tname] = {
                        "data": var.reshape((n1, n2), order="F"),
                        "unit": unit,
                    }
                else:
                    var = var.reshape((n1, n2, var.size // (n1 * n2)), order="F")
                    for i in range(var.shape[-1]):
                        tvals[_cool_species[i]] = {
                            "data": var[:, :, i],
                            "unit": "1/cm**3",
                        }

        # Add the mu field first, as it is needed for the number density
        interp = BilinearFieldInterpolator(
            tvals["mu"]["data"],
            (avals["lognH"], avals["logT"]),
            ["lognH", "logT"],
            truncate=True,
        )
        _create_field(("gas", "mu"), interp, "dimensionless")

        # Add the number density field, based on mu
        def _number_density(field, data):
            return data["gas", "density"] / mp / data["gas", "mu"]

        self.add_field(
            name=("gas", "number_density"),
            sampling_type="cell",
            function=_number_density,
            units=number_density_unit,
        )

        # Add the cooling and heating fields, which need the number density field
        for key in tvals:
            if key != "mu":
                interp = BilinearFieldInterpolator(
                    tvals[key]["data"],
                    (avals["lognH"], avals["logT"]),
                    ["lognH", "logT"],
                    truncate=True,
                )
                _create_field(("gas", key), interp, tvals[key]["unit"])

        # Add total cooling and heating fields
        def _all_cool(field, data):
            return (
                data["gas", "cooling_primordial"]
                + data["gas", "cooling_metal"]
                + data["gas", "cooling_compton"]
            )

        def _all_heat(field, data):
            return data["gas", "heating_primordial"] + data["gas", "heating_compton"]

        self.add_field(
            name=("gas", "cooling_total"),
            sampling_type="cell",
            function=_all_cool,
            units=cooling_function_units,
        )
        self.add_field(
            name=("gas", "heating_total"),
            sampling_type="cell",
            function=_all_heat,
            units=cooling_function_units,
        )

        # Add net cooling fields
        def _net_cool(field, data):
            return data["gas", "cooling_total"] - data["gas", "heating_total"]

        self.add_field(
            name=("gas", "cooling_net"),
            sampling_type="cell",
            function=_net_cool,
            units=cooling_function_units,
        )

        return True



from yt.testing import _geom_transforms, fake_amr_ds
from yt.utilities.answer_testing.framework import AxialPixelizationTest


def test_axial_pixelization():
    for geom in sorted(_geom_transforms):
        if geom == "spectral_cube":
            # skip this case as it was added much later and we don't want to keep
            # adding yield-based tests during the nose->pytest migration
            continue
        ds = fake_amr_ds(geometry=geom)
        yield AxialPixelizationTest(ds)



import numpy as np
from numpy.testing import (
    assert_array_equal,
    assert_array_less,
    assert_equal,
    assert_raises,
)

from yt.testing import fake_random_ds
from yt.utilities.lib.misc_utilities import (
    obtain_position_vector,
    obtain_relative_velocity_vector,
)

_fields = ("density", "velocity_x", "velocity_y", "velocity_z")
_units = ("g/cm**3", "cm/s", "cm/s", "cm/s")

# TODO: error compact/spread bits for incorrect size
# TODO: test msdb for [0,0], [1,1], [2,2] etc.


def test_spread_bits():
    from yt.utilities.lib.geometry_utils import spread_bits

    li = [
        (
            np.uint64(0b111111111111111111111),
            np.uint64(0b1001001001001001001001001001001001001001001001001001001001001),
        )
    ]
    for i, ans in li:
        out = spread_bits(i)
        assert_equal(out, ans)


def test_compact_bits():
    from yt.utilities.lib.geometry_utils import compact_bits

    li = [
        (
            np.uint64(0b111111111111111111111),
            np.uint64(0b1001001001001001001001001001001001001001001001001001001001001),
        )
    ]
    for ans, i in li:
        out = compact_bits(i)
        assert_equal(out, ans)


def test_spread_and_compact_bits():
    from yt.utilities.lib.geometry_utils import compact_bits, spread_bits

    li = [np.uint64(0b111111111111111111111)]
    for ans in li:
        mi = spread_bits(ans)
        out = compact_bits(mi)
        assert_equal(out, ans)


def test_lsz():
    from yt.utilities.lib.geometry_utils import lsz

    li = [
        (
            np.uint64(0b1001001001001001001001001001001001001001001001001001001001001),
            3 * 21,
            3,
            0,
        ),
        (
            np.uint64(0b1001001001001001001001001001001001001001001001001001001001000),
            3 * 0,
            3,
            0,
        ),
        (
            np.uint64(0b1001001001001001001001001001001001001001001001001001001000001),
            3 * 1,
            3,
            0,
        ),
        (
            np.uint64(0b1001001001001001001001001001001001001001001001001001000001001),
            3 * 2,
            3,
            0,
        ),
        (
            np.uint64(0b10010010010010010010010010010010010010010010010010010010010010),
            3 * 0,
            3,
            0,
        ),
        (
            np.uint64(
                0b100100100100100100100100100100100100100100100100100100100100100
            ),
            3 * 0,
            3,
            0,
        ),
        (np.uint64(0b100), 0, 1, 0),
        (np.uint64(0b100), 1, 1, 1),
        (np.uint64(0b100), 3, 1, 2),
        (np.uint64(0b100), 3, 1, 3),
    ]
    for i, ans, stride, start in li:
        out = lsz(i, stride=stride, start=start)
        assert_equal(out, ans)


def test_lsb():
    from yt.utilities.lib.geometry_utils import lsb

    li = [
        (
            np.uint64(0b1001001001001001001001001001001001001001001001001001001001001),
            3 * 0,
        ),
        (
            np.uint64(0b1001001001001001001001001001001001001001001001001001001001000),
            3 * 1,
        ),
        (
            np.uint64(0b1001001001001001001001001001001001001001001001001001001000000),
            3 * 2,
        ),
        (
            np.uint64(0b1001001001001001001001001001001001001001001001001001000000000),
            3 * 3,
        ),
        (
            np.uint64(0b10010010010010010010010010010010010010010010010010010010010010),
            3 * 21,
        ),
        (
            np.uint64(
                0b100100100100100100100100100100100100100100100100100100100100100
            ),
            3 * 21,
        ),
    ]
    for i, ans in li:
        out = lsb(i, stride=3)
        assert_equal(out, ans)


def test_bitwise_addition():
    from yt.utilities.lib.geometry_utils import bitwise_addition

    # TODO: Handle negative & periodic boundaries
    lz = [
        (0, 1),
        #          (0,-1),
        (1, 1),
        (1, 2),
        (1, 4),
        (1, -1),
        (2, 1),
        (2, 2),
        (2, -1),
        (2, -2),
        (3, 1),
        (3, 5),
        (3, -1),
    ]
    for i, a in lz:
        i = np.uint64(i)
        a = np.int64(a)
        out = bitwise_addition(i, a, stride=1, start=0)
        assert_equal(out, i + a)


# def test_add_to_morton_coord():
#    from yt.utilities.lib.geometry_utils import add_to_morton_coord


def test_get_morton_indices():
    from yt.utilities.lib.geometry_utils import (
        get_morton_indices,
        get_morton_indices_unravel,
    )

    INDEX_MAX_64 = np.uint64(2097151)
    li = np.arange(6, dtype=np.uint64).reshape((2, 3))
    mi_ans = np.array([10, 229], dtype=np.uint64)
    mi_out = get_morton_indices(li)
    mi_out2 = get_morton_indices_unravel(li[:, 0], li[:, 1], li[:, 2])
    assert_array_equal(mi_out, mi_ans)
    assert_array_equal(mi_out2, mi_ans)
    li[0, :] = INDEX_MAX_64 * np.ones(3, dtype=np.uint64)
    assert_raises(ValueError, get_morton_indices, li)
    assert_raises(ValueError, get_morton_indices_unravel, li[:, 0], li[:, 1], li[:, 2])


def test_get_morton_points():
    from yt.utilities.lib.geometry_utils import get_morton_points

    mi = np.array([10, 229], dtype=np.uint64)
    li_ans = np.arange(6, dtype=np.uint64).reshape((2, 3))
    li_out = get_morton_points(mi)
    assert_array_equal(li_out, li_ans)


def test_compare_morton():
    # TODO: Add error messages to assertions
    from yt.utilities.lib.geometry_utils import compare_morton

    # Diagonal
    p = np.array([0.0, 0.0, 0.0], dtype=np.float64)
    q = np.array([1.0, 1.0, 1.0], dtype=np.float64)
    assert_equal(compare_morton(p, q), 1)
    assert_equal(compare_morton(q, p), 0)
    assert_equal(compare_morton(p, p), 0)
    # 1-1 vs 0-1
    p = np.array([1.0, 1.0, 0.0], dtype=np.float64)
    q = np.array([1.0, 1.0, 1.0], dtype=np.float64)
    assert_equal(compare_morton(p, q), 1)
    assert_equal(compare_morton(q, p), 0)
    assert_equal(compare_morton(p, p), 0)
    # x advance, y decrease
    p = np.array([0.0, 1.0, 0.0], dtype=np.float64)
    q = np.array([1.0, 0.0, 0.0], dtype=np.float64)
    assert_equal(compare_morton(p, q), 1)
    assert_equal(compare_morton(q, p), 0)
    assert_equal(compare_morton(p, p), 0)
    # x&y advance, z decrease
    p = np.array([0.0, 0.0, 1.0], dtype=np.float64)
    q = np.array([1.0, 1.0, 0.0], dtype=np.float64)
    assert_equal(compare_morton(p, q), 1)
    assert_equal(compare_morton(q, p), 0)
    assert_equal(compare_morton(p, p), 0)


def test_get_morton_neighbors_coarse():
    from yt.utilities.lib.geometry_utils import get_morton_neighbors_coarse

    imax = 5
    ngz = 1
    tests = {
        (7, 1): np.array(
            [
                35,
                49,
                56,
                48,
                33,
                40,
                32,
                42,
                34,
                3,
                17,
                24,
                16,
                1,
                8,
                0,
                10,
                2,
                21,
                28,
                20,
                5,
                12,
                4,
                14,
                6,
            ],
            dtype="uint64",
        ),
        (7, 0): np.array(
            [
                35,
                49,
                56,
                48,
                33,
                40,
                32,
                42,
                34,
                3,
                17,
                24,
                16,
                1,
                8,
                0,
                10,
                2,
                21,
                28,
                20,
                5,
                12,
                4,
                14,
                6,
            ],
            dtype="uint64",
        ),
        (0, 1): np.array(
            [
                4,
                6,
                7,
                70,
                132,
                133,
                196,
                5,
                68,
                256,
                258,
                259,
                322,
                384,
                385,
                448,
                257,
                320,
                2,
                3,
                66,
                128,
                129,
                192,
                1,
                64,
            ],
            dtype="uint64",
        ),
        (0, 0): np.array([4, 6, 7, 5, 2, 3, 1], dtype="uint64"),
        (448, 1): np.array(
            [
                192,
                64,
                0,
                9,
                82,
                18,
                27,
                128,
                137,
                228,
                100,
                36,
                45,
                118,
                54,
                63,
                164,
                173,
                320,
                256,
                265,
                338,
                274,
                283,
                384,
                393,
            ],
            dtype="uint64",
        ),
        (448, 0): np.array([228, 118, 63, 173, 338, 283, 393], dtype="uint64"),
    }
    for (mi1, periodic), ans in tests.items():
        n1 = get_morton_neighbors_coarse(mi1, imax, periodic, ngz)
        assert_equal(np.sort(n1), np.sort(ans))


def test_get_morton_neighbors_refined():
    from yt.utilities.lib.geometry_utils import get_morton_neighbors_refined

    imax1 = 5
    imax2 = 5
    ngz = 1
    tests = {
        (7, 7, 1): (
            np.array(
                [
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                ],
                dtype="uint64",
            ),
            np.array(
                [
                    35,
                    49,
                    56,
                    48,
                    33,
                    40,
                    32,
                    42,
                    34,
                    3,
                    17,
                    24,
                    16,
                    1,
                    8,
                    0,
                    10,
                    2,
                    21,
                    28,
                    20,
                    5,
                    12,
                    4,
                    14,
                    6,
                ],
                dtype="uint64",
            ),
        ),
        (7, 7, 0): (
            np.array(
                [
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                    7,
                ],
                dtype="uint64",
            ),
            np.array(
                [
                    35,
                    49,
                    56,
                    48,
                    33,
                    40,
                    32,
                    42,
                    34,
                    3,
                    17,
                    24,
                    16,
                    1,
                    8,
                    0,
                    10,
                    2,
                    21,
                    28,
                    20,
                    5,
                    12,
                    4,
                    14,
                    6,
                ],
                dtype="uint64",
            ),
        ),
        (0, 0, 1): (
            np.array(
                [
                    0,
                    0,
                    0,
                    64,
                    128,
                    128,
                    192,
                    0,
                    64,
                    256,
                    256,
                    256,
                    320,
                    384,
                    384,
                    448,
                    256,
                    320,
                    0,
                    0,
                    64,
                    128,
                    128,
                    192,
                    0,
                    64,
                ],
                dtype="uint64",
            ),
            np.array(
                [
                    4,
                    6,
                    7,
                    70,
                    132,
                    133,
                    196,
                    5,
                    68,
                    256,
                    258,
                    259,
                    322,
                    384,
                    385,
                    448,
                    257,
                    320,
                    2,
                    3,
                    66,
                    128,
                    129,
                    192,
                    1,
                    64,
                ],
                dtype="uint64",
            ),
        ),
        (0, 0, 0): (
            np.array([0, 0, 0, 0, 0, 0, 0], dtype="uint64"),
            np.array([4, 6, 7, 5, 2, 3, 1], dtype="uint64"),
        ),
        (448, 448, 1): (
            np.array(
                [
                    192,
                    64,
                    0,
                    64,
                    192,
                    128,
                    192,
                    128,
                    192,
                    448,
                    320,
                    256,
                    320,
                    448,
                    384,
                    448,
                    384,
                    448,
                    320,
                    256,
                    320,
                    448,
                    384,
                    448,
                    384,
                    448,
                ],
                dtype="uint64",
            ),
            np.array(
                [
                    192,
                    64,
                    0,
                    9,
                    82,
                    18,
                    27,
                    128,
                    137,
                    228,
                    100,
                    36,
                    45,
                    118,
                    54,
                    63,
                    164,
                    173,
                    320,
                    256,
                    265,
                    338,
                    274,
                    283,
                    384,
                    393,
                ],
                dtype="uint64",
            ),
        ),
        (448, 448, 0): (
            np.array([448, 448, 448, 448, 448, 448, 448], dtype="uint64"),
            np.array([228, 118, 63, 173, 338, 283, 393], dtype="uint64"),
        ),
    }
    for (mi1, mi2, periodic), (ans1, ans2) in tests.items():
        n1, n2 = get_morton_neighbors_refined(mi1, mi2, imax1, imax2, periodic, ngz)
        assert_equal(np.sort(n1), np.sort(ans1))
        assert_equal(np.sort(n2), np.sort(ans2))


def test_morton_neighbor():
    from yt.utilities.lib.geometry_utils import get_morton_indices, morton_neighbor

    order = 20
    imax = np.uint64(1 << order)
    p = np.array(
        [
            [imax / 2, imax / 2, imax / 2],
            [imax / 2, imax / 2, 0],
            [imax / 2, imax / 2, imax],
        ],
        dtype=np.uint64,
    )
    p_ans = np.array(
        [
            [imax / 2, imax / 2, imax / 2 + 1],
            [imax / 2, imax / 2, imax / 2 - 1],
            [imax / 2, imax / 2, imax - 1],
            [imax / 2, imax / 2, 1],
            [imax / 2, imax / 2 + 1, imax / 2 + 1],
            [imax / 2 - 1, imax / 2 - 1, imax / 2],
            [imax / 2 - 1, imax / 2, imax / 2 + 1],
            [imax / 2, imax / 2 - 1, imax - 1],
            [imax / 2, imax / 2 + 1, 1],
        ],
        dtype=np.uint64,
    )
    mi_ans = get_morton_indices(p_ans)
    assert_equal(morton_neighbor(p[0, :], [2], [+1], imax), mi_ans[0])
    assert_equal(morton_neighbor(p[0, :], [2], [-1], imax), mi_ans[1])
    assert_equal(morton_neighbor(p[1, :], [2], [-1], imax, periodic=False), -1)
    assert_equal(morton_neighbor(p[2, :], [2], [+1], imax, periodic=False), -1)
    assert_equal(morton_neighbor(p[1, :], [2], [-1], imax, periodic=True), mi_ans[2])
    assert_equal(morton_neighbor(p[2, :], [2], [+1], imax, periodic=True), mi_ans[3])
    assert_equal(morton_neighbor(p[0, :], [1, 2], [+1, +1], imax), mi_ans[4])
    assert_equal(morton_neighbor(p[0, :], [0, 1], [-1, -1], imax), mi_ans[5])
    assert_equal(morton_neighbor(p[0, :], [0, 2], [-1, +1], imax), mi_ans[6])
    assert_equal(morton_neighbor(p[1, :], [1, 2], [-1, -1], imax, periodic=False), -1)
    assert_equal(morton_neighbor(p[2, :], [1, 2], [+1, +1], imax, periodic=False), -1)
    assert_equal(
        morton_neighbor(p[1, :], [1, 2], [-1, -1], imax, periodic=True), mi_ans[7]
    )
    assert_equal(
        morton_neighbor(p[2, :], [1, 2], [+1, +1], imax, periodic=True), mi_ans[8]
    )


def test_get_morton_neighbors():
    from yt.utilities.lib.geometry_utils import get_morton_indices, get_morton_neighbors

    order = 20
    imax = 1 << order
    p = np.array(
        [
            [imax / 2, imax / 2, imax / 2],
            [imax / 2, imax / 2, 0],
            [imax / 2, imax / 2, imax],
        ],
        dtype=np.uint64,
    )
    pn_non = [
        np.array(
            [
                # x +/- 1
                [imax / 2 + 1, imax / 2, imax / 2],
                [imax / 2 + 1, imax / 2 + 1, imax / 2],
                [imax / 2 + 1, imax / 2 + 1, imax / 2 + 1],
                [imax / 2 + 1, imax / 2 + 1, imax / 2 - 1],
                [imax / 2 + 1, imax / 2 - 1, imax / 2],
                [imax / 2 + 1, imax / 2 - 1, imax / 2 + 1],
                [imax / 2 + 1, imax / 2 - 1, imax / 2 - 1],
                [imax / 2 + 1, imax / 2, imax / 2 + 1],
                [imax / 2 + 1, imax / 2, imax / 2 - 1],
                [imax / 2 - 1, imax / 2, imax / 2],
                [imax / 2 - 1, imax / 2 + 1, imax / 2],
                [imax / 2 - 1, imax / 2 + 1, imax / 2 + 1],
                [imax / 2 - 1, imax / 2 + 1, imax / 2 - 1],
                [imax / 2 - 1, imax / 2 - 1, imax / 2],
                [imax / 2 - 1, imax / 2 - 1, imax / 2 + 1],
                [imax / 2 - 1, imax / 2 - 1, imax / 2 - 1],
                [imax / 2 - 1, imax / 2, imax / 2 + 1],
                [imax / 2 - 1, imax / 2, imax / 2 - 1],
                # y +/- 1
                [imax / 2, imax / 2 + 1, imax / 2],
                [imax / 2, imax / 2 + 1, imax / 2 + 1],
                [imax / 2, imax / 2 + 1, imax / 2 - 1],
                [imax / 2, imax / 2 - 1, imax / 2],
                [imax / 2, imax / 2 - 1, imax / 2 + 1],
                [imax / 2, imax / 2 - 1, imax / 2 - 1],
                # x +/- 1
                [imax / 2, imax / 2, imax / 2 + 1],
                [imax / 2, imax / 2, imax / 2 - 1],
            ],
            dtype=np.uint64,
        ),
        np.array(
            [
                # x +/- 1
                [imax / 2 + 1, imax / 2, 0],
                [imax / 2 + 1, imax / 2 + 1, 0],
                [imax / 2 + 1, imax / 2 + 1, 1],
                [imax / 2 + 1, imax / 2 - 1, 0],
                [imax / 2 + 1, imax / 2 - 1, 1],
                [imax / 2 + 1, imax / 2, 1],
                [imax / 2 - 1, imax / 2, 0],
                [imax / 2 - 1, imax / 2 + 1, 0],
                [imax / 2 - 1, imax / 2 + 1, 1],
                [imax / 2 - 1, imax / 2 - 1, 0],
                [imax / 2 - 1, imax / 2 - 1, 1],
                [imax / 2 - 1, imax / 2, 1],
                # y +/- 1
                [imax / 2, imax / 2 + 1, 0],
                [imax / 2, imax / 2 + 1, 1],
                [imax / 2, imax / 2 - 1, 0],
                [imax / 2, imax / 2 - 1, 1],
                # z +/- 1
                [imax / 2, imax / 2, 0 + 1],
            ],
            dtype=np.uint64,
        ),
        np.array(
            [
                # x +/- 1
                [imax / 2 + 1, imax / 2, imax],
                [imax / 2 + 1, imax / 2 + 1, imax],
                [imax / 2 + 1, imax / 2 + 1, imax - 1],
                [imax / 2 + 1, imax / 2 - 1, imax],
                [imax / 2 + 1, imax / 2 - 1, imax - 1],
                [imax / 2 + 1, imax / 2, imax - 1],
                [imax / 2 - 1, imax / 2, imax],
                [imax / 2 - 1, imax / 2 + 1, imax],
                [imax / 2 - 1, imax / 2 + 1, imax - 1],
                [imax / 2 - 1, imax / 2 - 1, imax],
                [imax / 2 - 1, imax / 2 - 1, imax - 1],
                [imax / 2 - 1, imax / 2, imax - 1],
                # y +/- 1
                [imax / 2, imax / 2 + 1, imax],
                [imax / 2, imax / 2 + 1, imax - 1],
                [imax / 2, imax / 2 - 1, imax],
                [imax / 2, imax / 2 - 1, imax - 1],
                # z +/- 1
                [imax / 2, imax / 2, imax - 1],
            ],
            dtype=np.uint64,
        ),
    ]
    pn_per = [
        np.array(
            [
                # x +/- 1
                [imax / 2 + 1, imax / 2, imax / 2],
                [imax / 2 + 1, imax / 2 + 1, imax / 2],
                [imax / 2 + 1, imax / 2 + 1, imax / 2 + 1],
                [imax / 2 + 1, imax / 2 + 1, imax / 2 - 1],
                [imax / 2 + 1, imax / 2 - 1, imax / 2],
                [imax / 2 + 1, imax / 2 - 1, imax / 2 + 1],
                [imax / 2 + 1, imax / 2 - 1, imax / 2 - 1],
                [imax / 2 + 1, imax / 2, imax / 2 + 1],
                [imax / 2 + 1, imax / 2, imax / 2 - 1],
                [imax / 2 - 1, imax / 2, imax / 2],
                [imax / 2 - 1, imax / 2 + 1, imax / 2],
                [imax / 2 - 1, imax / 2 + 1, imax / 2 + 1],
                [imax / 2 - 1, imax / 2 + 1, imax / 2 - 1],
                [imax / 2 - 1, imax / 2 - 1, imax / 2],
                [imax / 2 - 1, imax / 2 - 1, imax / 2 + 1],
                [imax / 2 - 1, imax / 2 - 1, imax / 2 - 1],
                [imax / 2 - 1, imax / 2, imax / 2 + 1],
                [imax / 2 - 1, imax / 2, imax / 2 - 1],
                # y +/- 1
                [imax / 2, imax / 2 + 1, imax / 2],
                [imax / 2, imax / 2 + 1, imax / 2 + 1],
                [imax / 2, imax / 2 + 1, imax / 2 - 1],
                [imax / 2, imax / 2 - 1, imax / 2],
                [imax / 2, imax / 2 - 1, imax / 2 + 1],
                [imax / 2, imax / 2 - 1, imax / 2 - 1],
                # z +/- 1
                [imax / 2, imax / 2, imax / 2 + 1],
                [imax / 2, imax / 2, imax / 2 - 1],
            ],
            dtype=np.uint64,
        ),
        np.array(
            [
                # x +/- 1
                [imax / 2 + 1, imax / 2, 0],
                [imax / 2 + 1, imax / 2 + 1, 0],
                [imax / 2 + 1, imax / 2 + 1, 1],
                [imax / 2 + 1, imax / 2 + 1, imax - 1],
                [imax / 2 + 1, imax / 2 - 1, 0],
                [imax / 2 + 1, imax / 2 - 1, 1],
                [imax / 2 + 1, imax / 2 - 1, imax - 1],
                [imax / 2 + 1, imax / 2, 1],
                [imax / 2 + 1, imax / 2, imax - 1],
                [imax / 2 - 1, imax / 2, 0],
                [imax / 2 - 1, imax / 2 + 1, 0],
                [imax / 2 - 1, imax / 2 + 1, 1],
                [imax / 2 - 1, imax / 2 + 1, imax - 1],
                [imax / 2 - 1, imax / 2 - 1, 0],
                [imax / 2 - 1, imax / 2 - 1, 1],
                [imax / 2 - 1, imax / 2 - 1, imax - 1],
                [imax / 2 - 1, imax / 2, 1],
                [imax / 2 - 1, imax / 2, imax - 1],
                # y +/- 1
                [imax / 2, imax / 2 + 1, 0],
                [imax / 2, imax / 2 + 1, 1],
                [imax / 2, imax / 2 + 1, imax - 1],
                [imax / 2, imax / 2 - 1, 0],
                [imax / 2, imax / 2 - 1, 1],
                [imax / 2, imax / 2 - 1, imax - 1],
                # z +/- 1
                [imax / 2, imax / 2, 0 + 1],
                [imax / 2, imax / 2, imax - 1],
            ],
            dtype=np.uint64,
        ),
        np.array(
            [
                # x +/- 1
                [imax / 2 + 1, imax / 2, imax],
                [imax / 2 + 1, imax / 2 + 1, imax],
                [imax / 2 + 1, imax / 2 + 1, 1],
                [imax / 2 + 1, imax / 2 + 1, imax - 1],
                [imax / 2 + 1, imax / 2 - 1, imax],
                [imax / 2 + 1, imax / 2 - 1, 1],
                [imax / 2 + 1, imax / 2 - 1, imax - 1],
                [imax / 2 + 1, imax / 2, 1],
                [imax / 2 + 1, imax / 2, imax - 1],
                [imax / 2 - 1, imax / 2, imax],
                [imax / 2 - 1, imax / 2 + 1, imax],
                [imax / 2 - 1, imax / 2 + 1, 1],
                [imax / 2 - 1, imax / 2 + 1, imax - 1],
                [imax / 2 - 1, imax / 2 - 1, imax],
                [imax / 2 - 1, imax / 2 - 1, 1],
                [imax / 2 - 1, imax / 2 - 1, imax - 1],
                [imax / 2 - 1, imax / 2, 1],
                [imax / 2 - 1, imax / 2, imax - 1],
                # y +/- 1
                [imax / 2, imax / 2 + 1, imax],
                [imax / 2, imax / 2 + 1, 1],
                [imax / 2, imax / 2 + 1, imax - 1],
                [imax / 2, imax / 2 - 1, imax],
                [imax / 2, imax / 2 - 1, 1],
                [imax / 2, imax / 2 - 1, imax - 1],
                # z +/- 1
                [imax / 2, imax / 2, 1],
                [imax / 2, imax / 2, imax - 1],
            ],
            dtype=np.uint64,
        ),
    ]
    mi = get_morton_indices(p)
    N = mi.shape[0]
    # Non-periodic
    for i in range(N):
        out = get_morton_neighbors(
            np.array([mi[i]], dtype=np.uint64), order=order, periodic=False
        )
        ans = get_morton_indices(np.vstack([p[i, :], pn_non[i]]))
        assert_array_equal(np.unique(out), np.unique(ans), err_msg=f"Non-periodic: {i}")
    # Periodic
    for i in range(N):
        out = get_morton_neighbors(
            np.array([mi[i]], dtype=np.uint64), order=order, periodic=True
        )
        ans = get_morton_indices(np.vstack([p[i, :], pn_per[i]]))
        assert_array_equal(np.unique(out), np.unique(ans), err_msg=f"Periodic: {i}")


def test_dist():
    from yt.utilities.lib.geometry_utils import dist

    p = np.array([0.0, 0.0, 0.0], dtype=np.float64)
    q = np.array([0.0, 0.0, 0.0], dtype=np.float64)
    assert_equal(dist(p, q), 0.0)
    p = np.array([0.0, 0.0, 0.0], dtype=np.float64)
    q = np.array([1.0, 0.0, 0.0], dtype=np.float64)
    assert_equal(dist(p, q), 1.0)
    p = np.array([0.0, 0.0, 0.0], dtype=np.float64)
    q = np.array([1.0, 1.0, 0.0], dtype=np.float64)
    assert_equal(dist(p, q), np.sqrt(2.0))
    p = np.array([0.0, 0.0, 0.0], dtype=np.float64)
    q = np.array([1.0, 1.0, 1.0], dtype=np.float64)
    assert_equal(dist(p, q), np.sqrt(3.0))


def test_knn_direct(seed=1):
    from yt.utilities.lib.geometry_utils import knn_direct

    np.random.seed(seed)
    k = 64
    N = 1e5
    idx = np.arange(N, dtype=np.uint64)
    rad = np.arange(N, dtype=np.float64)
    pos = np.vstack(3 * [rad**2 / 3.0]).T
    sort_shf = np.arange(N, dtype=np.uint64)
    for _ in range(20):
        np.random.shuffle(sort_shf)
        sort_ans = np.argsort(sort_shf)[:k]
        sort_out = knn_direct(pos[sort_shf, :], k, sort_ans[0], idx)
        assert_array_equal(sort_out, sort_ans)


# TODO: test of quadtree (.pxd)


def test_obtain_position_vector():
    ds = fake_random_ds(
        64, nprocs=8, fields=_fields, units=_units, negative=[False, True, True, True]
    )

    dd = ds.sphere((0.5, 0.5, 0.5), 0.2)

    coords = obtain_position_vector(dd)

    r = np.sqrt(np.sum(coords * coords, axis=0))

    assert_array_less(r.max(), 0.2)

    assert_array_less(0.0, r.min())


def test_obtain_relative_velocity_vector():
    ds = fake_random_ds(
        64, nprocs=8, fields=_fields, units=_units, negative=[False, True, True, True]
    )

    dd = ds.all_data()

    vels = obtain_relative_velocity_vector(dd)

    assert_array_equal(vels[0, :], dd["gas", "velocity_x"])
    assert_array_equal(vels[1, :], dd["gas", "velocity_y"])
    assert_array_equal(vels[2, :], dd["gas", "velocity_z"])



import yt

ds = yt.load_sample("IsolatedGalaxy")

fields = [
    ("gas", "density"),
    ("gas", "velocity_x"),
    ("gas", "velocity_y"),
    ("gas", "velocity_magnitude"),
]

p = yt.SlicePlot(ds, "z", fields)
p.set_log(("gas", "velocity_x"), False)
p.set_log(("gas", "velocity_y"), False)

# this returns a matplotlib figure with an ImageGrid and the slices
# added to the grid of axes (in this case, 2x2)
fig = p.export_to_mpl_figure((2, 2))

fig.tight_layout()

fig.savefig("multiplot_export_to_mpl.png")



import numpy as np
import pytest
from numpy.testing import assert_array_equal

from yt.testing import fake_particle_ds, requires_module
from yt.utilities.exceptions import YTFieldNotFound


@requires_module("firefly")
def test_firefly_JSON_string():
    ds = fake_particle_ds()
    ad = ds.all_data()
    reader = ad.create_firefly_object(
        None,
        velocity_units="cm/s",
        coordinate_units="cm",
    )

    reader.writeToDisk(write_to_disk=False, file_extension=".json")

    ## reader.JSON was not output to string correctly
    ##  either Firefly is damaged or needs a hotfix-- try reinstalling.
    ##  if that doesn't work contact the developers
    ##  at github.com/ageller/Firefly/issues.
    assert len(reader.JSON) > 0


@requires_module("firefly")
def test_firefly_write_to_disk(tmp_path):
    tmpdir = str(tmp_path)  # create_firefly_object needs a str, not PosixPath

    ds = fake_particle_ds()
    ad = ds.all_data()
    reader = ad.create_firefly_object(
        tmpdir,
        velocity_units="cm/s",
        coordinate_units="cm",
        match_any_particle_types=True,  # Explicitly specifying to avoid deprecation warning
    )

    reader.writeToDisk()


@pytest.fixture
def firefly_test_dataset():
    # create dataset
    ds_fields = [
        # Assumed present
        ("pt1", "particle_position_x"),
        ("pt1", "particle_position_y"),
        ("pt1", "particle_position_z"),
        ("pt2", "particle_position_x"),
        ("pt2", "particle_position_y"),
        ("pt2", "particle_position_z"),
        # User input
        ("pt1", "common_field"),
        ("pt2", "common_field"),
        ("pt2", "pt2only_field"),
    ]
    ds_field_units = ["code_length"] * 9
    ds_negative = [0] * 9
    ds = fake_particle_ds(
        fields=ds_fields,
        units=ds_field_units,
        negative=ds_negative,
    )
    return ds


@requires_module("firefly")
@pytest.mark.parametrize(
    "fields_to_include,fields_units",
    [
        (None, None),  # Test default values
        ([], []),  # Test empty fields
    ],
)
def test_field_empty_specification(
    firefly_test_dataset, fields_to_include, fields_units
):
    dd = firefly_test_dataset.all_data()
    reader = dd.create_firefly_object(
        fields_to_include=fields_to_include,
        fields_units=fields_units,
        coordinate_units="code_length",
    )
    assert_array_equal(
        dd["pt1", "relative_particle_position"].d,
        reader.particleGroups[0].coordinates,
    )
    assert_array_equal(
        dd["pt2", "relative_particle_position"].d,
        reader.particleGroups[1].coordinates,
    )


@requires_module("firefly")
def test_field_unique_string_specification(firefly_test_dataset):
    # Test unique field (pt2only_field)
    dd = firefly_test_dataset.all_data()
    # Unique field string will fallback to "all" field type and fail
    # as nonexistent ("all", "pt2only_field") unless we set
    # match_any_particle_types=True
    reader = dd.create_firefly_object(
        fields_to_include=["pt2only_field"],
        fields_units=["code_length"],
        coordinate_units="code_length",
        match_any_particle_types=True,
    )

    pt1 = reader.particleGroups[0]
    pt2 = reader.particleGroups[1]
    assert_array_equal(
        dd["pt1", "relative_particle_position"].d,
        pt1.coordinates,
    )
    assert_array_equal(
        dd["pt2", "relative_particle_position"].d,
        pt2.coordinates,
    )
    assert "pt2only_field" not in pt1.field_names
    assert "pt2only_field" in pt2.field_names
    arrind = np.flatnonzero(pt2.field_names == "pt2only_field")[0]
    assert_array_equal(dd["pt2", "pt2only_field"].d, pt2.field_arrays[arrind])


@requires_module("firefly")
def test_field_common_string_specification(firefly_test_dataset):
    # Test common field (common_field)
    dd = firefly_test_dataset.all_data()
    # Common field string will be ambiguous and fail
    # unless we set match_any_particle_types=True
    reader = dd.create_firefly_object(
        fields_to_include=["common_field"],
        fields_units=["code_length"],
        coordinate_units="code_length",
        match_any_particle_types=True,
    )

    pt1 = reader.particleGroups[0]
    pt2 = reader.particleGroups[1]
    assert_array_equal(
        dd["pt1", "relative_particle_position"].d,
        pt1.coordinates,
    )
    assert_array_equal(
        dd["pt2", "relative_particle_position"].d,
        pt2.coordinates,
    )
    assert "common_field" in pt1.field_names
    assert "common_field" in pt2.field_names
    arrind = np.flatnonzero(pt1.field_names == "common_field")[0]
    assert_array_equal(dd["pt1", "common_field"].d, pt1.field_arrays[arrind])
    arrind = np.flatnonzero(pt2.field_names == "common_field")[0]
    assert_array_equal(dd["pt2", "common_field"].d, pt2.field_arrays[arrind])


@requires_module("firefly")
@pytest.mark.parametrize(
    "fields_to_include,fields_units",
    [
        (
            [("pt2", "pt2only_field")],
            ["code_length"],
        ),  # Test existing field tuple (pt2, pt2only_field)
        (
            [("pt1", "common_field")],
            ["code_length"],
        ),  # Test that tuples only bring in referenced particleGroup
        (
            [("all", "common_field")],
            ["code_length"],
        ),  # Test that "all" brings in all particleGroups
    ],
)
def test_field_tuple_specification(
    firefly_test_dataset,
    fields_to_include,
    fields_units,
):
    dd = firefly_test_dataset.all_data()
    reader = dd.create_firefly_object(
        fields_to_include=fields_to_include,
        fields_units=fields_units,
        coordinate_units="code_length",
    )
    assert_array_equal(
        dd["pt1", "relative_particle_position"].d,
        reader.particleGroups[0].coordinates,
    )
    assert_array_equal(
        dd["pt2", "relative_particle_position"].d,
        reader.particleGroups[1].coordinates,
    )
    all_pgs = reader.particleGroups
    all_pgs_names = ["pt1", "pt2"]
    for field in fields_to_include:
        ftype, fname = field
        for pgi in range(2):
            pg = all_pgs[pgi]
            if ftype == all_pgs_names[pgi]:
                assert fname in pg.field_names
                arrind = np.flatnonzero(pg.field_names == fname)[0]
                assert_array_equal(dd[field].d, pg.field_arrays[arrind])
            elif ftype == "all":
                assert fname in pg.field_names
                this_pg_name = all_pgs_names[pgi]
                arrind = np.flatnonzero(pg.field_names == fname)[0]
                assert_array_equal(dd[this_pg_name, fname].d, pg.field_arrays[arrind])
            else:
                assert fname not in pg.field_names


@requires_module("firefly")
@pytest.mark.parametrize(
    "fields_to_include,fields_units,ErrorType",
    [
        (
            ["dinos"],
            ["code_length"],
            YTFieldNotFound,
        ),  # Test nonexistent field (dinos)
        (
            ["common_field"],
            ["code_length"],
            ValueError,
        ),  # Test ambiguous field (match_any_particle_types=False)
        (
            [("pt1", "pt2only_field")],
            ["code_length"],
            YTFieldNotFound,
        ),  # Test nonexistent field tuple (pt1, pt2only_field)
    ],
)
def test_field_invalid_specification(
    firefly_test_dataset, fields_to_include, fields_units, ErrorType
):
    dd = firefly_test_dataset.all_data()
    # Note that we have specified match_any_particle_types as False since
    # that is the behavior expected in the future
    with pytest.raises(ErrorType):
        dd.create_firefly_object(
            fields_to_include=fields_to_include,
            fields_units=fields_units,
            coordinate_units="code_length",
            match_any_particle_types=False,
        )


@requires_module("firefly")
def test_field_mixed_specification(firefly_test_dataset):
    dd = firefly_test_dataset.all_data()

    reader = dd.create_firefly_object(
        fields_to_include=["pt2only_field", ("pt1", "common_field")],
        fields_units=["code_length", "code_length"],
    )

    pt1 = reader.particleGroups[0]
    pt2 = reader.particleGroups[1]
    assert "common_field" in pt1.field_names
    assert "common_field" not in pt2.field_names
    arrind = np.flatnonzero(pt1.field_names == "common_field")[0]
    assert_array_equal(dd["pt1", "common_field"].d, pt1.field_arrays[arrind])

    assert "pt2only_field" not in pt1.field_names
    assert "pt2only_field" in pt2.field_names
    arrind = np.flatnonzero(pt2.field_names == "pt2only_field")[0]
    assert_array_equal(dd["pt2", "pt2only_field"].d, pt2.field_arrays[arrind])



# distutils: libraries = STD_LIBS
"""
CyOctree building, loading and refining routines
"""
cimport cython
cimport libc.math as math
cimport numpy as np

import numpy as np

from libc.stdlib cimport free, malloc

from yt.geometry.particle_deposit cimport get_kernel_func, kernel_func

np.import_array()


cdef struct Octree:
    # Array of 3*num_nodes [x1, y1, z1, x2, y2, z2, ...]
    np.float64_t * node_positions
    # 1 or 0 of whether the oct has refined to make children
    np.uint8_t * refined
    # Each oct stores the depth in the tree: the root node is 0
    np.uint8_t * depth
    # pstart and pend tell us which particles in the pidx are stored in each oct
    np.int64_t * pstart
    np.int64_t * pend
    # This tells us the index of each child
    np.int64_t * children

    # Here we store the coordinates and IDs of all the particles in the tree
    np.float64_t * pposx
    np.float64_t * pposy
    np.float64_t * pposz
    np.int64_t * pidx

    # The max number of particles per leaf, if above, we refine
    np.int64_t n_ref

    # The number of particles in our tree, e.g the length of ppos and pidx
    np.int64_t num_particles

    # The total size of the octree (x, y, z)
    np.float64_t * size

    # The current number of nodes in the octree
    np.int64_t num_nodes

    # The maximum depth before we stop refining, and the maximum number of nodes
    # we can fit in our array
    np.uint8_t max_depth
    np.int64_t max_num_nodes


@cython.boundscheck(False)
@cython.wraparound(False)
cdef int octree_build_node(Octree * tree, long int node_idx):
    """
    This is the main function in the building of the octree. This function takes
    in the tree and an index of a oct to process. If the node has too many
    particles (> n_ref) and the depth is less than the maximum tree depth then
    we refine.

    The `refine` creates 8 sub octs, within our oct. We indenfity the particles
    in each of the subocts. We then recursively call this function on each of
    the subocts.

    Parameters
    ----------
    tree : Octree *
        A pointer to the octree
    node_idx : long int
        The index of the current node we are processing

    Returns
    -------
    int
        Success of tree build
    """
    cdef np.int64_t splits[9]
    cdef np.int64_t i, j, k, n, start, end
    cdef np.float64_t lx, ly, lz, sz, inv_size

    # If we are running out of space in our tree, then we *try* to
    # relloacate a tree of double the size
    if (tree.num_nodes + 8) >= tree.max_num_nodes:
        if octree_reallocate(tree, tree.max_num_nodes * 2):
            return 1

    if (
        (tree.pend[node_idx] - tree.pstart[node_idx] > tree.n_ref) and
        (tree.depth[node_idx] < tree.max_depth)
    ):
        tree.refined[node_idx] = 1

        # As we have decided to refine, we need to know which of the particles
        # in this oct will go into each of the 8 child octs
        split_helper(tree, node_idx, splits)

        # Figure out the size of the current oct
        inv_size = 1. / 2.**tree.depth[node_idx]
        sx = tree.size[0] * inv_size
        sy = tree.size[1] * inv_size
        sz = tree.size[2] * inv_size
        lx = tree.node_positions[3*node_idx] - sx/2.
        ly = tree.node_positions[(3*node_idx)+1] - sy/2.
        lz = tree.node_positions[(3*node_idx)+2] - sz/2.

        # Loop through and generate the children AND recursively refine...
        n = 0
        for i in range(2):
            for j in range(2):
                for k in range(2):
                    start = splits[n]
                    end = splits[n + 1]
                    child = tree.num_nodes

                    # Store the child location
                    tree.children[8*node_idx + n] = child

                    tree.node_positions[child*3] = lx + sx*i
                    tree.node_positions[(child*3)+1] = ly + sy*j
                    tree.node_positions[(child*3)+2] = lz + sz*k

                    tree.refined[child] = 0
                    tree.depth[child] = tree.depth[node_idx] + 1

                    tree.pstart[child] = start
                    tree.pend[child] = end
                    tree.num_nodes += 1

                    # Recursively refine child
                    if octree_build_node(tree, child):
                        return 1
                    n += 1

    return 0


@cython.boundscheck(False)
@cython.wraparound(False)
cdef int octree_allocate(Octree * octree, long int num_nodes):
    """
    This is the main allocation function in the octree. We allocate all of the
    arrays we require to store information about every single oct in the tree

    Parameters
    ----------
    octree : Octree *
        A pointer to the octree
    num_nodes : long int
        The maximum number of nodes to allocate for

    Returns
    -------
    int
        Success of allocation
    """
    octree.node_positions = <np.float64_t *> malloc(
        num_nodes * 3 * sizeof(np.float64_t))
    if octree.node_positions == NULL:
        return 1

    octree.size = <np.float64_t *> malloc(3 * sizeof(np.float64_t))
    if octree.size == NULL:
        return 1

    octree.children = <np.int64_t *> malloc(8 * num_nodes * sizeof(np.int64_t))
    if octree.children == NULL:
        return 1

    octree.pstart = <np.int64_t *> malloc(num_nodes * sizeof(np.int64_t))
    if octree.pstart == NULL:
        return 1

    octree.pend = <np.int64_t *> malloc(num_nodes * sizeof(np.int64_t))
    if octree.pend == NULL:
        return 1

    octree.refined = <np.uint8_t *> malloc(num_nodes * sizeof(np.int8_t))
    if octree.refined == NULL:
        return 1

    octree.depth = <np.uint8_t *> malloc(num_nodes * sizeof(np.int8_t))
    if octree.depth == NULL:
        return 1

    return 0


@cython.boundscheck(False)
@cython.wraparound(False)
cdef int octree_reallocate(Octree * octree, long int num_nodes):
    """
    This function re-allocates all of the arrays malloc'd in `octree_allocate`
    See Notes for when we want to re-allocate.

    Parameters
    ----------
    octree : Octree *
        A pointer to the octree
    num_nodes : long int
        The maximum number of nodes to (re)allocate for

    Returns
    -------
    int
        Success of the reallocation

    Notes
    -----
    Why do we want to re-allocate?

    Well 2 cases,
    1) The octree is still building and we have ran out of space, so we
    have asked for an increased number of nodes and are reallocating each array

    2) We have finished building the octree and we have used less nodes
    than we originally allocated. We are now shrinking the octree and giving
    the spare memory back.
    """
    cdef np.float64_t * old_arr
    cdef np.int64_t * old_arr_int
    cdef np.uint8_t * old_arr_uint
    cdef np.int64_t i

    old_arr = octree.node_positions
    octree.node_positions = <np.float64_t *> malloc(num_nodes * 3 * sizeof(np.float64_t))
    if octree.node_positions == NULL: return 1
    for i in range(3*octree.num_nodes):
        octree.node_positions[i] = old_arr[i]
    free(old_arr)

    old_arr_int = octree.children
    octree.children = <np.int64_t *> malloc(num_nodes * 8 * sizeof(np.int64_t))
    if octree.children == NULL: return 1
    for i in range(8*octree.num_nodes):
        octree.children[i] = old_arr_int[i]
    free(old_arr_int)

    old_arr_int = octree.pstart
    octree.pstart = <np.int64_t *> malloc(num_nodes * sizeof(np.int64_t))
    if octree.pstart == NULL: return 1
    for i in range(octree.num_nodes):
        octree.pstart[i] = old_arr_int[i]
    free(old_arr_int)

    old_arr_int = octree.pend
    octree.pend = <np.int64_t *> malloc(num_nodes * sizeof(np.int64_t))
    if octree.pend == NULL: return 1
    for i in range(octree.num_nodes):
        octree.pend[i] = old_arr_int[i]
    free(old_arr_int)

    old_arr_uint = octree.refined
    octree.refined = <np.uint8_t *> malloc(num_nodes * sizeof(np.int8_t))
    if octree.refined == NULL: return 1
    for i in range(octree.num_nodes):
        octree.refined[i] = old_arr_uint[i]
    free(old_arr_uint)

    old_arr_uint = octree.depth
    octree.depth = <np.uint8_t *> malloc(num_nodes * sizeof(np.int8_t))
    if octree.depth == NULL: return 1
    for i in range(octree.num_nodes):
        octree.depth[i] = old_arr_uint[i]
    free(old_arr_uint)

    octree.max_num_nodes = num_nodes

    return 0


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void octree_deallocate(Octree * octree):
    """
    Just free-ing every array we allocated to ensure we don't leak.

    Parameter
    ---------
    octree : Octree *
        Pointer to the octree
    """
    free(octree.node_positions)
    free(octree.size)
    free(octree.children)

    free(octree.pstart)
    free(octree.pend)

    free(octree.refined)
    free(octree.depth)

    free(octree.pidx)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef class CyOctree:
    """
    This a class to store the underlying octree and particle data that can be
    interacted with from both Cython and Python
    """
    cdef Octree * c_octree
    cdef np.float64_t[::1, :] input_positions

    cdef np.int64_t n_ref
    cdef np.float64_t[:] left_edge
    cdef np.float64_t[:] right_edge
    cdef np.uint8_t max_depth

    cdef kernel_func kernel

    def __init__(
        self,
        np.float64_t[:, :] input_pos,
        left_edge=None,
        right_edge=None,
        np.int64_t n_ref=32,
        np.uint8_t max_depth=200
    ):
        """
        Octree initialiser.

        We copy the inputted particle positions and make the root node. We then
        refine the octree until every leaf has either less particles than n_ref
        or is at the maximum depth.

        Finally, we re-allocate all of the memory required by tree to ensure we
        do not use more memory than required.

        Parameters
        ----------
        input_pos : 2D memory view
            Particles positions in the format (num_particles, 3)
        {left,right}_edge : ndarray
            xyz coordinates of the lower left (upper right) corner of the octree.
        n_ref : int, default: 32
            The maximum number of particles per leaf, if more, the oct
            will refine
        max_depth : int, default: 200
            The maximum depth the octree will refine to. If we set
            this too high then we may hit a stack overflow due to the
            recursive nature of the build
        """
        self.n_ref = n_ref
        self.max_depth = max_depth
        self.input_positions = np.asfortranarray(input_pos, dtype=np.float64)

        if self._allocate_octree():
            raise MemoryError("Unable to allocate memory required for octree build.")

        self._make_root(left_edge, right_edge)

        if octree_build_node(self.c_octree, 0):
            raise MemoryError("Unable to allocate memory required for octree build.")

        if octree_reallocate(self.c_octree, self.c_octree.num_nodes):
            raise MemoryError("Unable to allocate memory required for octree build.")

    def __del__(self):
        """
        Make sure we clean up properly!
        """
        octree_deallocate(self.c_octree)
        free(self.c_octree)

    @property
    def bound_particles(self):
        """
        The particle selection may select SPH particles with smoothing lengths
        which is in the octree domains. However, if the particle center is NOT
        in the octree, they are not included.

        So the number of particles passed to the tree *may* not be equal to the
        number of particles which are bound by the tree.
        """
        return self.c_octree.pend[0]

    @property
    def num_nodes(self):
        """
        The total number of nodes after tree construction
        """
        return self.c_octree.num_nodes

    @property
    def node_positions(self):
        """
        The centre of every node within the octree
        """
        cdef np.npy_intp shape[2]
        shape[0] = <np.npy_intp> self.c_octree.num_nodes
        shape[1] = 3
        arr = np.PyArray_SimpleNewFromData(
            2, &shape[0], np.NPY_FLOAT64, <void *>self.c_octree.node_positions)
        return np.copy(arr)

    @property
    def node_refined(self):
        """
        An array of length num_nodes which contains either True / False for
        whether each cell has refined or not. E.g False for a leaf, True for a
        node
        """
        cdef np.npy_intp shape
        shape = <np.npy_intp> self.c_octree.num_nodes
        arr = np.PyArray_SimpleNewFromData(
            1, &shape, np.NPY_UINT8, <void *>self.c_octree.refined)
        return np.copy(arr).astype(np.bool_)

    @property
    def node_depth(self):
        """
        The depth for each node in the tree. The root node is defined as a
        depth of 0.
        """
        cdef np.npy_intp shape = <np.npy_intp> self.c_octree.num_nodes
        arr = np.PyArray_SimpleNewFromData(
            1, &shape, np.NPY_UINT8, <void *>self.c_octree.depth)
        return np.copy(arr)

    @property
    def node_sizes(self):
        """
        The size of each node in the x, y and z directions. We calculate this
        on the fly. As we know the size of the whole tree and the depth of each
        node
        """
        cdef np.int64_t i
        sizes = np.zeros((self.c_octree.num_nodes, 3), dtype=np.float64)
        sizes[:, 0] = self.c_octree.size[0]
        sizes[:, 1] = self.c_octree.size[1]
        sizes[:, 2] = self.c_octree.size[2]

        for i in range(self.c_octree.num_nodes):
            sizes[i, :] /= <np.float64_t> (2.0**self.c_octree.depth[i] / 2.0)

        return sizes

    def _make_root(self, left_edge, right_edge):
        """
        The root node is the hardest node to build as we need to find out which
        particles we contain, and then sieving them into children is easy.

        In the case that the left_edge/right_edge is not defined then we select
        a tree that is sufficiently big to contain every particle.

        However, if they are defined then we need to loop through and find out
        which particles are actually in the tree.

        Parameters
        ----------
        {left,right}_edge : ndarray
            xyz coordinates of the lower left (upper right) corner of the octree.
            If None, the tree will be made large enough to encompass all particles.
        """
        cdef int i = 0

        # How many particles are there?
        self.c_octree.num_particles = self.input_positions.shape[0]

        # We now number all of the the particles in the tree. This allows us
        # to shuffle the pids and say for example Oct11 contains particles
        # 7 to 11
        # This pidx[7:12] would give the input indices of the particles we
        # store. This proxy allows us to re-arrange the particles without
        # re-arranging the users input data.
        self.c_octree.pidx = <np.int64_t *> malloc(
            self.c_octree.num_particles * sizeof(np.int64_t)
        )
        for i in range(0, self.c_octree.num_particles):
            self.c_octree.pidx[i] = i

        if left_edge is None:
            # If the edges are None, then we can just find the loop through
            # and find them out
            left_edge = np.zeros(3, dtype=np.float64)
            right_edge = np.zeros(3, dtype=np.float64)

            left_edge[0] = self.c_octree.pposx[0]
            left_edge[1] = self.c_octree.pposy[0]
            left_edge[2] = self.c_octree.pposz[0]
            right_edge[0] = self.c_octree.pposx[0]
            right_edge[1] = self.c_octree.pposy[0]
            right_edge[2] = self.c_octree.pposz[0]

            for i in range(self.c_octree.num_particles):
                left_edge[0] = min(self.c_octree.pposx[i], left_edge[0])
                left_edge[1] = min(self.c_octree.pposy[i], left_edge[1])
                left_edge[2] = min(self.c_octree.pposz[i], left_edge[2])

                right_edge[0] = max(self.c_octree.pposx[i], right_edge[0])
                right_edge[1] = max(self.c_octree.pposy[i], right_edge[1])
                right_edge[2] = max(self.c_octree.pposz[i], right_edge[2])

            self.c_octree.pstart[0] = 0
            self.c_octree.pend[0] = self.input_positions.shape[0]
        else:
            # Damn! The user did supply a left and right so we need to find
            # which particles are in the range
            left_edge = left_edge.astype(np.float64)
            right_edge = right_edge.astype(np.float64)

            # We loop through the particles and arrange them such that particles
            # in the tree are to the left of the split and the particles not
            # are to the right
            # e.g.
            # pidx = [1, 2, 3, 5 | 0, 4]
            # where split = 4 and particles 0 and 4 are not in the tree
            split = select(
                self.c_octree, left_edge, right_edge, 0, self.input_positions.shape[0])
            self.c_octree.pstart[0] = 0
            self.c_octree.pend[0] = split

        # Set the total size of the tree
        size = (right_edge - left_edge) / 2.0
        center = (right_edge + left_edge) / 2.0
        self.left_edge = left_edge
        self.right_edge = right_edge

        # Now we add the data about the root node!
        self.c_octree.node_positions[0] = center[0]
        self.c_octree.node_positions[1] = center[1]
        self.c_octree.node_positions[2] = center[2]

        self.c_octree.size[0] = size[0]
        self.c_octree.size[1] = size[1]
        self.c_octree.size[2] = size[2]

        # We are not refined yet
        self.c_octree.refined[0] = 0
        self.c_octree.depth[0] = 0

    def _allocate_octree(self):
        """
        This actually allocates the C struct Octree
        """
        self.c_octree = <Octree*> malloc(sizeof(Octree))
        self.c_octree.n_ref = self.n_ref
        self.c_octree.num_nodes = 1

        # This is sort of an arbitrary guess but it doesn't matter because
        # we will increase this value and attempt to reallocate if it is too
        # small
        self.c_octree.max_num_nodes = max(self.input_positions.shape[0] / self.n_ref, 1)
        self.c_octree.max_depth = self.max_depth

        # Fast C pointers to the particle coordinates
        self.c_octree.pposx = &self.input_positions[0, 0]
        self.c_octree.pposy = &self.input_positions[0, 1]
        self.c_octree.pposz = &self.input_positions[0, 2]

        if octree_allocate(self.c_octree, self.c_octree.max_num_nodes): return 1

    cdef void smooth_onto_cells(
        self,
        np.float64_t[:] buff,
        np.float64_t[:] buff_den,
        np.float64_t posx,
        np.float64_t posy,
        np.float64_t posz,
        np.float64_t hsml,
        np.float64_t prefactor,
        np.float64_t prefactor_norm,
        long int num_node,
        int use_normalization=0
    ):
        """
        We smooth a field onto cells within an octree using SPH deposition. To
        achieve this we loop through every oct in the tree and check if it is a
        leaf. A leaf just means that an oct has not refined, and thus has no
        children.

        Parameters
        ----------
        buff : memoryview
            The array which we are depositing the field onto, it has the
            length of the number of leaves.
        buff_den : memoryview
            The array we deposit just mass onto to allow normalization
        pos<> : float64_t
            The x, y, and z coordinates of the particle we are depositing
        hsml : float64_t
            The smoothing length of the particle
        prefactor(_norm) : float64_t
            This is a pre-computed value, based on the particles
                properties used in the deposition
        num_node : lonmg int
            The current node we are checking to see if refined or not
        use_normalization : int, default: 0
            Do we want a normalized sph field? If so, fill the buff_den.
        """

        cdef Octree * tree = self.c_octree
        cdef double q_ij, diff_x, diff_y, diff_z, diff, sx, sy, sz
        cdef int i
        cdef long int child_node

        if tree.refined[num_node] == 0:
            diff_x = tree.node_positions[3*num_node] - posx
            diff_y = tree.node_positions[3*num_node+1] - posy
            diff_z = tree.node_positions[3*num_node+2] - posz

            q_ij = math.sqrt(diff_x*diff_x + diff_y*diff_y + diff_z*diff_z)
            q_ij /= hsml

            buff[num_node] += (prefactor * self.kernel(q_ij))
            if use_normalization:
                buff_den[num_node] += (prefactor_norm * self.kernel(q_ij))

        else:
            # All direct children of the current node are the same size, thus
            # we can compute their size once, outside of the loop
            sz_factor = 1.0 / 2.0**(tree.depth[num_node] + 1)
            sqrt_sz_factor = math.sqrt(sz_factor)
            sx = tree.size[0]
            sy = tree.size[1]
            sz = tree.size[2]
            child_node_size = sqrt_sz_factor * math.sqrt(sx*sx + sy*sy + sz*sz)

            for i in range(8):
                child_node = tree.children[8*num_node + i]
                diff_x = tree.node_positions[3*child_node] - posx
                diff_y = tree.node_positions[3*child_node+1] - posy
                diff_z = tree.node_positions[3*child_node+2] - posz
                diff = math.sqrt(diff_x*diff_x + diff_y*diff_y + diff_z*diff_z)

                # Could the current particle possibly intersect this child node?
                if diff - child_node_size - hsml < 0:
                    self.smooth_onto_cells(buff, buff_den, posx, posy, posz,
                                    hsml, prefactor, prefactor_norm,
                                    child_node, use_normalization=use_normalization)


    def interpolate_sph_cells(self,
        np.float64_t[:] buff,
        np.float64_t[:] buff_den,
        np.float64_t[:] posx,
        np.float64_t[:] posy,
        np.float64_t[:] posz,
        np.float64_t[:] pmass,
        np.float64_t[:] pdens,
        np.float64_t[:] hsml,
        np.float64_t[:] field,
        kernel_name="cubic",
        int use_normalization=0
    ):
        """
        We loop through every particle in the simulation and begin to deposit
        the particle properties onto all of the leaves that it intersects

        Parameters
        ----------
        buff : memoryview
            The array which we are depositing the field onto, it has the
            length of the number of leaves.
        buff_den : memoryview
            The array we deposit just mass onto to allow normalization
        pos<> : memoryview
            The x, y, and z coordinates of all the partciles
        pmass : memoryview
            The mass of the particles
        pdens : memoryview
            The density of the particles
        hsml : memoryview
            The smoothing lengths of the particles
        field : memoryview
            The field we are depositing for each particle
        kernel_name: str, default: "cubic"
            Choice of kernel for SPH deposition
        use_normalization : int, default: 0
            Do we want a normalized sph field? If so, fill the buff_den.
        """
        self.kernel = get_kernel_func(kernel_name)

        cdef int i
        cdef double prefactor, prefactor_norm
        for i in range(posx.shape[0]):
            prefactor = pmass[i] / pdens[i] / hsml[i]**3
            prefactor_norm = prefactor
            prefactor *= field[i]

            self.smooth_onto_cells(buff, buff_den, posx[i], posy[i], posz[i],
                                    hsml[i], prefactor, prefactor_norm,
                                    0, use_normalization=use_normalization)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef np.int64_t separate(
        np.float64_t * pos,
        np.int64_t * pidx,
        double value,
        np.int64_t start,
        np.int64_t end
    ) noexcept nogil:
    """
    This is a simple utility function which takes a selection of particles and
    re-arranges them such that values below `value` are to the left of split and
    values above are to the right.

    Parameters
    ----------
    pos : float64_t *
        Pointer to the coordinates we are splitting along
    pidx : int64_t &
        Pointer to the corresponding particle IDs
    value : double
        The value to split the data along
    start : int64_t
        Index of first particle in the current node
    end : int64_t
        Index of the last particle in the current node
    """
    cdef np.int64_t index
    cdef np.int64_t split = start
    for index in range(start, end):
        if pos[pidx[index]] < value:
            pidx[split], pidx[index] = pidx[index], pidx[split]
            split +=  1

    return split


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void split_helper(Octree * tree, np.int64_t node_idx, np.int64_t * splits):
    """
    A utility function to split a collection of particles along the x, y and z
    direction such that we identify the particles within the 8 children of an
    oct.

    We first split the particles in the x direction, then we split the particles
    in the y and the z directions. This is currently hardcoded for 8 octs but
    can be readily extended to allow over-refinement.

    The splits[0] and splits[1] tell oct one the start and last particle
    The splits[1] and splits[2] tell oct two the start and last particle and so
    on

    Parameters
    ----------
    tree : Octree *
        A pointer to the octree
    node_idx :  int64_t
        The index of the oct we are splitting
    splits : int64_t *
        Pointer to split array which stores the start and end indices. It needs
        to be N+1 long where N is the number of children
    """
    splits[0] = tree.pstart[node_idx]
    splits[8] = tree.pend[node_idx]

    splits[4] = separate(
        tree.pposx, tree.pidx, tree.node_positions[3*node_idx], splits[0], splits[8])

    splits[2] = separate(
        tree.pposy, tree.pidx, tree.node_positions[(3*node_idx)+1], splits[0], splits[4])
    splits[6] = separate(
        tree.pposy, tree.pidx, tree.node_positions[(3*node_idx)+1], splits[4], splits[8])

    splits[1] = separate(
        tree.pposz, tree.pidx, tree.node_positions[(3*node_idx)+2], splits[0], splits[2])
    splits[3] = separate(
        tree.pposz, tree.pidx, tree.node_positions[(3*node_idx)+2], splits[2], splits[4])
    splits[5] = separate(
        tree.pposz, tree.pidx, tree.node_positions[(3*node_idx)+2], splits[4], splits[6])
    splits[7] = separate(
        tree.pposz, tree.pidx, tree.node_positions[(3*node_idx)+2], splits[6], splits[8])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef np.int64_t select(
        Octree * octree,
        np.float64_t[::1] left_edge,
        np.float64_t[::1] right_edge,
        np.int64_t start,
        np.int64_t end
    ) noexcept nogil:
    """
    Re-arrange the input particles such that those outside the bounds of the
    tree occur after the split index and can thus be ignored for the remainder
    of the tree construction

    Parameters
    ----------
    tree : Octree *
        A pointer to the octree
    left_edge : ndarray
        The coords of the lower left corner of the box
    right_edge : ndarray
        The coords of the upper right corner of the box
    start : int64_t
        The first particle in the bounds (0)
    end : int64_t
        The last particle in the bounds
    """
    cdef np.int64_t index
    cdef np.int64_t split = start

    cdef np.float64_t * posx = octree.pposx
    cdef np.float64_t * posy = octree.pposy
    cdef np.float64_t * posz = octree.pposz
    cdef np.int64_t * pidx = octree.pidx

    for index in range(start, end):
        if posx[pidx[index]] < right_edge[0] and posx[pidx[index]] > left_edge[0]:
            if posy[pidx[index]] < right_edge[1] and posy[pidx[index]] > left_edge[1]:
                if posz[pidx[index]] < right_edge[2] and posz[pidx[index]] > left_edge[2]:
                    if split < index:
                        pidx[split], pidx[index] = pidx[index], pidx[split]
                    split += 1

    return split



import yt
from yt.visualization.volume_rendering.api import Scene, create_volume_source

filePath = "Sedov_3d/sedov_hdf5_chk_0003"
ds = yt.load(filePath)
ds.force_periodicity()

sc = Scene()

# set up camera
cam = sc.add_camera(ds, lens_type="perspective")
cam.resolution = [400, 400]

cam.position = ds.arr([1, 1, 1], "cm")
cam.switch_orientation()

# add rendering of density field
dens = create_volume_source(ds, field=("flash", "dens"))
dens.use_ghost_zones = True
sc.add_source(dens)
sc.save("density.png", sigma_clip=6)

# add rendering of x-velocity field
vel = create_volume_source(ds, field=("flash", "velx"))
vel.use_ghost_zones = True
sc.add_source(vel)
sc.save("density_any_velocity.png", sigma_clip=6)



import numpy as np

import yt
from yt.testing import (
    assert_rel_equal,
    cubicspline_python,
    fake_sph_flexible_grid_ds,
    integrate_kernel,
    requires_file,
)
from yt.utilities.math_utils import compute_stddev_image

## off-axis projection tests for SPH data are in
## yt/visualization/tests/test_offaxisprojection.py


magneticum = "MagneticumCluster/snap_132"

mag_kwargs = {
    "long_ids": True,
    "field_spec": "magneticum_box2_hr",
}


@requires_file(magneticum)
def test_sph_moment():
    ds = yt.load(magneticum, **mag_kwargs)

    def _vysq(field, data):
        return data["gas", "velocity_y"] ** 2

    ds.add_field(("gas", "vysq"), _vysq, sampling_type="local", units="cm**2/s**2")
    prj1 = yt.ProjectionPlot(
        ds,
        "y",
        [("gas", "velocity_y"), ("gas", "vysq")],
        weight_field=("gas", "density"),
        moment=1,
        buff_size=(400, 400),
    )
    prj2 = yt.ProjectionPlot(
        ds,
        "y",
        ("gas", "velocity_y"),
        moment=2,
        weight_field=("gas", "density"),
        buff_size=(400, 400),
    )
    sigy = compute_stddev_image(prj1.frb["gas", "vysq"], prj1.frb["gas", "velocity_y"])
    assert_rel_equal(sigy, prj2.frb["gas", "velocity_y"].d, 10)


def test_sph_projection_basic1():
    """
    small, uniform grid: expected values for given dl?
    pixel centers at 0.5, 1., 1.5, 2., 2.5
    particles at 0.5, 1.5, 2.5
    """
    bbox = np.array([[0.0, 3.0]] * 3)
    ds = fake_sph_flexible_grid_ds(hsml_factor=1.0, nperside=3, bbox=bbox)
    # works, but no depth control (at least without specific filters)
    proj = ds.proj(("gas", "density"), 2)
    frb = proj.to_frb(
        width=(2.5, "cm"),
        resolution=(5, 5),
        height=(2.5, "cm"),
        center=np.array([1.5, 1.5, 1.5]),
        periodic=False,
    )
    out = frb.get_image(("gas", "density"))

    expected_out = np.zeros((5, 5), dtype=np.float64)
    dl_1part = integrate_kernel(cubicspline_python, 0.0, 0.5)
    linedens_1part = dl_1part * 1.0  # unit mass, density
    linedens = 3.0 * linedens_1part
    expected_out[::2, ::2] = linedens

    assert_rel_equal(expected_out, out.v, 5)
    # return out


def test_sph_projection_basic2():
    """
    small, uniform grid: expected values for given dl?
    pixel centers at 0.5, 1., 1.5, 2., 2.5
    particles at 0.5, 1.5, 2.5
    but hsml radii are 0.25 -> try non-zero impact parameters,
    other pixels are still zero.
    """
    bbox = np.array([[0.0, 3.0]] * 3)
    ds = fake_sph_flexible_grid_ds(hsml_factor=0.5, nperside=3, bbox=bbox)
    proj = ds.proj(("gas", "density"), 2)
    frb = proj.to_frb(
        width=(2.5, "cm"),
        resolution=(5, 5),
        height=(2.5, "cm"),
        center=np.array([1.375, 1.375, 1.5]),
        periodic=False,
    )
    out = frb.get_image(("gas", "density"))

    expected_out = np.zeros((5, 5), dtype=np.float64)
    dl_1part = integrate_kernel(cubicspline_python, np.sqrt(2) * 0.125, 0.25)
    linedens_1part = dl_1part * 1.0  # unit mass, density
    linedens = 3.0 * linedens_1part
    expected_out[::2, ::2] = linedens

    # print(expected_out)
    # print(out.v)
    assert_rel_equal(expected_out, out.v, 4)
    # return out


def get_dataset_sphrefine(reflevel: int = 1):
    """
    constant density particle grid,
    with increasing particle sampling
    """
    lenfact = (1.0 / 3.0) ** (reflevel - 1)
    massfact = lenfact**3
    nperside = 3**reflevel

    e1hat = np.array([lenfact, 0, 0])
    e2hat = np.array([0, lenfact, 0])
    e3hat = np.array([0, 0, lenfact])
    hsml_factor = lenfact
    bbox = np.array([[0.0, 3.0]] * 3)
    offsets = np.ones(3, dtype=np.float64) * 0.5  # in units of ehat

    def refmass(i: int, j: int, k: int) -> float:
        return massfact

    unitrho = 1.0 / massfact  # want density 1 for decreasing mass

    ds = fake_sph_flexible_grid_ds(
        hsml_factor=hsml_factor,
        nperside=nperside,
        periodic=True,
        e1hat=e1hat,
        e2hat=e2hat,
        e3hat=e3hat,
        offsets=offsets,
        massgenerator=refmass,
        unitrho=unitrho,
        bbox=bbox,
    )
    return ds


def getdata_test_gridproj2():
    # initial pixel centers at 0.5, 1., 1.5, 2., 2.5
    # particles at 0.5, 1.5, 2.5
    # refine particle grid, check if pixel values remain the
    # same in the pixels passing through initial particle centers
    outlist = []
    dss = []
    for rl in range(1, 4):
        ds = get_dataset_sphrefine(reflevel=rl)
        proj = ds.proj(("gas", "density"), 2)
        frb = proj.to_frb(
            width=(2.5, "cm"),
            resolution=(5, 5),
            height=(2.5, "cm"),
            center=np.array([1.5, 1.5, 1.5]),
            periodic=False,
        )
        out = frb.get_image(("gas", "density"))
        outlist.append(out)
        dss.append(ds)
    return outlist, dss


def test_sph_gridproj_reseffect1():
    """
    Comparing same pixel centers with higher particle resolution.
    The pixel centers are at x/y coordinates [0.5, 1., 1.5, 2., 2.5]
    at the first level, the spacing halves at each level.
    Checking the pixels at [0.5, 1.5, 2.5],
    which should have the same values at each resolution.
    """
    imgs, _ = getdata_test_gridproj2()
    ref = imgs[-1]
    for img in imgs:
        assert_rel_equal(
            img[:: img.shape[0] // 2, :: img.shape[1] // 2],
            ref[:: ref.shape[0] // 2, :: ref.shape[1] // 2],
            4,
        )


def test_sph_gridproj_reseffect2():
    """
    refine the pixel grid instead of the particle grid
    """
    ds = get_dataset_sphrefine(reflevel=2)
    proj = ds.proj(("gas", "density"), 2)
    imgs = {}
    maxrl = 5
    for rl in range(1, maxrl + 1):
        npix = 1 + 2 ** (rl + 1)
        margin = 0.5 - 0.5 ** (rl + 1)
        frb = proj.to_frb(
            width=(3.0 - 2.0 * margin, "cm"),
            resolution=(npix, npix),
            height=(3.0 - 2.0 * margin, "cm"),
            center=np.array([1.5, 1.5, 1.5]),
            periodic=False,
        )
        out = frb.get_image(("gas", "density"))
        imgs[rl] = out
    ref = imgs[maxrl]
    pixspace_ref = 2 ** (maxrl)
    for rl in imgs:
        img = imgs[rl]
        pixspace = 2 ** (rl)
        # print(f'Grid refinement level {rl}:')
        assert_rel_equal(
            img[::pixspace, ::pixspace], ref[::pixspace_ref, ::pixspace_ref], 4
        )



import os

import numpy as np

from yt.data_objects.static_output import ParticleDataset
from yt.funcs import mylog
from yt.geometry.particle_geometry_handler import ParticleIndex


class SPHDataset(ParticleDataset):
    default_kernel_name = "cubic"
    _sph_smoothing_styles = ["scatter", "gather"]
    _sph_smoothing_style = "scatter"
    _num_neighbors = 32
    _use_sph_normalization = True

    def __init__(
        self,
        filename,
        dataset_type=None,
        units_override=None,
        unit_system="cgs",
        index_order=None,
        index_filename=None,
        kdtree_filename=None,
        kernel_name=None,
        default_species_fields=None,
    ):
        if kernel_name is None:
            self.kernel_name = self.default_kernel_name
        else:
            self.kernel_name = kernel_name
        self.kdtree_filename = kdtree_filename
        super().__init__(
            filename,
            dataset_type=dataset_type,
            units_override=units_override,
            unit_system=unit_system,
            index_order=index_order,
            index_filename=index_filename,
            default_species_fields=default_species_fields,
        )

    @property
    def num_neighbors(self):
        return self._num_neighbors

    @num_neighbors.setter
    def num_neighbors(self, value):
        if value < 0:
            raise ValueError(f"Negative value not allowed: {value}")
        self._num_neighbors = value

    @property
    def sph_smoothing_style(self):
        return self._sph_smoothing_style

    @sph_smoothing_style.setter
    def sph_smoothing_style(self, value):
        if value not in self._sph_smoothing_styles:
            raise ValueError(
                f"Smoothing style not implemented: {value}, "
                "please select one of the following: ",
                self._sph_smoothing_styles,
            )

        self._sph_smoothing_style = value

    @property
    def use_sph_normalization(self):
        return self._use_sph_normalization

    @use_sph_normalization.setter
    def use_sph_normalization(self, value):
        if value is not True and value is not False:
            raise ValueError("SPH normalization needs to be True or False!")
        self._use_sph_normalization = value


class SPHParticleIndex(ParticleIndex):
    def _initialize_index(self):
        ds = self.dataset

        ds._file_hash = self._generate_hash()

        if hasattr(self.io, "_generate_smoothing_length"):
            self.io._generate_smoothing_length(self)

        super()._initialize_index()

    def _generate_kdtree(self, fname):
        from yt.utilities.lib.cykdtree import PyKDTree

        if fname is not None:
            if os.path.exists(fname):
                mylog.info("Loading KDTree from %s", os.path.basename(fname))
                kdtree = PyKDTree.from_file(fname)
                if kdtree.data_version != self.ds._file_hash:
                    mylog.info("Detected hash mismatch, regenerating KDTree")
                else:
                    self._kdtree = kdtree
                    return
        positions = []
        for data_file in self.data_files:
            for _, ppos in self.io._yield_coordinates(
                data_file, needed_ptype=self.ds._sph_ptypes[0]
            ):
                positions.append(ppos)
        if positions == []:
            self._kdtree = None
            return
        positions = np.concatenate(positions)
        mylog.info("Allocating KDTree for %s particles", positions.shape[0])
        num_neighbors = getattr(self.ds, "num_neighbors", 32)
        self._kdtree = PyKDTree(
            positions.astype("float64"),
            left_edge=self.ds.domain_left_edge,
            right_edge=self.ds.domain_right_edge,
            periodic=np.array(self.ds.periodicity),
            leafsize=2 * int(num_neighbors),
            data_version=self.ds._file_hash,
        )
        if fname is not None:
            self._kdtree.save(fname)

    @property
    def kdtree(self):
        if hasattr(self, "_kdtree"):
            return self._kdtree

        ds = self.ds

        if getattr(ds, "kdtree_filename", None) is None:
            if os.path.exists(ds.parameter_filename):
                fname = ds.parameter_filename + ".kdtree"
            else:
                # we don't want to write to disk for in-memory data
                fname = None
        else:
            fname = ds.kdtree_filename

        self._generate_kdtree(fname)

        return self._kdtree



import os
import shutil
import tempfile

import numpy as np
from numpy.testing import assert_array_equal

from yt.loaders import load, load_uniform_grid
from yt.testing import (
    assert_fname,
    fake_random_ds,
    requires_file,
    requires_module,
)
from yt.utilities.answer_testing.framework import data_dir_load
from yt.visualization.plot_window import ProjectionPlot, SlicePlot

ytdata_dir = "ytdata_test"


@requires_module("h5py")
@requires_file(os.path.join(ytdata_dir, "slice.h5"))
@requires_file(os.path.join(ytdata_dir, "proj.h5"))
@requires_file(os.path.join(ytdata_dir, "oas.h5"))
def test_old_plot_data():
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    fn = "slice.h5"
    full_fn = os.path.join(ytdata_dir, fn)
    ds_slice = data_dir_load(full_fn)
    p = SlicePlot(ds_slice, "z", ("gas", "density"))
    fn = p.save()
    assert_fname(fn[0])

    fn = "proj.h5"
    full_fn = os.path.join(ytdata_dir, fn)
    ds_proj = data_dir_load(full_fn)
    p = ProjectionPlot(ds_proj, "z", ("gas", "density"))
    fn = p.save()
    assert_fname(fn[0])

    fn = "oas.h5"
    full_fn = os.path.join(ytdata_dir, fn)
    ds_oas = data_dir_load(full_fn)
    p = SlicePlot(ds_oas, [1, 1, 1], ("gas", "density"))
    fn = p.save()
    assert_fname(fn[0])

    os.chdir(curdir)
    shutil.rmtree(tmpdir)


@requires_module("h5py")
def test_plot_data():
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)
    ds = fake_random_ds(16)

    plot = SlicePlot(ds, "z", ("gas", "density"))
    fn = plot.data_source.save_as_dataset("slice.h5")
    ds_slice = load(fn)
    p = SlicePlot(ds_slice, "z", ("gas", "density"))
    fn = p.save()
    assert_fname(fn[0])

    plot = ProjectionPlot(ds, "z", ("gas", "density"))
    fn = plot.data_source.save_as_dataset("proj.h5")
    ds_proj = load(fn)
    p = ProjectionPlot(ds_proj, "z", ("gas", "density"))
    fn = p.save()
    assert_fname(fn[0])

    plot = SlicePlot(ds, [1, 1, 1], ("gas", "density"))
    fn = plot.data_source.save_as_dataset("oas.h5")
    ds_oas = load(fn)
    p = SlicePlot(ds_oas, [1, 1, 1], ("gas", "density"))
    fn = p.save()
    assert_fname(fn[0])

    os.chdir(curdir)
    if tmpdir != ".":
        shutil.rmtree(tmpdir)


@requires_module("h5py")
def test_non_square_frb():
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    # construct an arbitrary dataset
    arr = np.arange(8.0 * 9.0 * 10.0).reshape((8, 9, 10))
    data = {"density": (arr, "g/cm**3")}
    bbox = np.array([[-4, 4.0], [-4.5, 4.5], [-5.0, 5]])
    ds = load_uniform_grid(
        data, arr.shape, length_unit="Mpc", bbox=bbox, periodicity=(False, False, False)
    )

    # make a slice
    slc = ds.slice(axis="z", coord=ds.quan(0.0, "code_length"))
    # make a frb and save it to disk
    center = (ds.quan(0.0, "code_length"), ds.quan(0.0, "code_length"))
    xax, yax = ds.coordinates.x_axis[slc.axis], ds.coordinates.y_axis[slc.axis]
    res = [ds.domain_dimensions[xax], ds.domain_dimensions[yax]]  # = [8,9]
    width = ds.domain_right_edge[xax] - ds.domain_left_edge[xax]  # = 8 code_length
    height = ds.domain_right_edge[yax] - ds.domain_left_edge[yax]  # = 9 code_length
    frb = slc.to_frb(width=width, height=height, resolution=res, center=center)
    fname = "test_frb_roundtrip.h5"
    frb.save_as_dataset(fname, fields=[("gas", "density")])

    expected_vals = arr[:, :, 5].T
    print(
        "\nConfirmation that initial frb results are expected:",
        (expected_vals == frb["gas", "density"].v).all(),
        "\n",
    )

    # yt-reload:
    reloaded_ds = load(fname)

    assert_array_equal(
        frb["gas", "density"].shape, reloaded_ds.data["gas", "density"].shape
    )
    assert_array_equal(frb["gas", "density"], reloaded_ds.data["gas", "density"])

    os.chdir(curdir)
    if tmpdir != ".":
        shutil.rmtree(tmpdir)



from yt.frontends.sph.fields import SPHFieldInfo


class SwiftFieldInfo(SPHFieldInfo):
    def __init__(self, ds, field_list, slice_info=None):
        self.known_particle_fields += (
            (
                "InternalEnergies",
                ("code_specific_energy", ["specific_thermal_energy"], None),
            ),
            ("Densities", ("code_mass / code_length**3", ["density"], None)),
            ("SmoothingLengths", ("code_length", ["smoothing_length"], None)),
        )
        super().__init__(ds, field_list, slice_info)

    def setup_particle_fields(self, ptype, *args, **kwargs):
        super().setup_particle_fields(ptype, *args, **kwargs)

        if ptype in ("PartType0", "Gas"):
            self.setup_gas_particle_fields(ptype)

    def setup_gas_particle_fields(self, ptype):
        self.alias((ptype, "temperature"), (ptype, "Temperatures"))
        self.alias(("gas", "temperature"), (ptype, "Temperatures"))

        for ax in ("x", "y", "z"):
            self.alias((ptype, ax), (ptype, "particle_position_" + ax))



import inspect

from yt.utilities.object_registries import analysis_task_registry


class AnalysisTask:
    def __init_subclass__(cls, *args, **kwargs):
        if hasattr(cls, "skip") and not cls.skip:
            return
        analysis_task_registry[cls.__name__] = cls

    def __init__(self, *args, **kwargs):
        # This should only get called if the subclassed object
        # does not override
        if len(args) + len(kwargs) != len(self._params):
            raise RuntimeError
        self.__dict__.update(zip(self._params, args, strict=False))
        self.__dict__.update(kwargs)

    def __repr__(self):
        # Stolen from YTDataContainer.__repr__
        s = f"{self.__class__.__name__}: "
        s += ", ".join(f"{i}={getattr(self, i)}" for i in self._params)
        return s


def analysis_task(params=None):
    if params is None:
        params = ()

    def create_new_class(func):
        cls = type(func.__name__, (AnalysisTask,), {"eval": func, "_params": params})
        return cls

    return create_new_class


@analysis_task(("field",))
def MaximumValue(params, data_object):
    v = data_object.quantities["MaxLocation"](params.field)[0]
    return v


@analysis_task()
def CurrentTimeYears(params, ds):
    return ds.current_time * ds["years"]


class SlicePlotDataset(AnalysisTask):
    _params = ["field", "axis", "center"]

    def __init__(self, *args, **kwargs):
        from yt.visualization.api import SlicePlot

        self.SlicePlot = SlicePlot
        AnalysisTask.__init__(self, *args, **kwargs)

    def eval(self, ds):
        slc = self.SlicePlot(ds, self.axis, self.field, center=self.center)
        return slc.save()


class QuantityProxy(AnalysisTask):
    _params = None
    quantity_name = None

    def __repr__(self):
        # Stolen from YTDataContainer.__repr__
        s = f"{self.__class__.__name__}: "
        s += ", ".join([str(list(self.args))])
        s += ", ".join(f"{k}={v}" for k, v in self.kwargs.items())
        return s

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def eval(self, data_object):
        rv = data_object.quantities[self.quantity_name](*self.args, **self.kwargs)
        return rv


class ParameterValue(AnalysisTask):
    _params = ["parameter"]

    def __init__(self, parameter, cast=None):
        self.parameter = parameter
        if cast is None:

            def _identity(x):
                return x

            cast = _identity
        self.cast = cast

    def eval(self, ds):
        return self.cast(ds.get_parameter(self.parameter))


def create_quantity_proxy(quantity_object):
    args, varargs, kwargs, defaults = inspect.getargspec(quantity_object[1])
    # Strip off 'data' which is on every quantity function
    params = args[1:]
    if kwargs is not None:
        params += kwargs
    dd = {"_params": params, "quantity_name": quantity_object[0]}
    cls = type(quantity_object[0], (QuantityProxy,), dd)
    return cls



import numpy as np

from yt.data_objects.selection_objects.data_selection_objects import (
    YTSelectionContainer,
    YTSelectionContainer2D,
)
from yt.data_objects.static_output import Dataset
from yt.funcs import (
    fix_length,
    is_sequence,
    iter_fields,
    validate_3d_array,
    validate_axis,
    validate_center,
    validate_float,
    validate_object,
    validate_width_tuple,
)
from yt.utilities.exceptions import YTNotInsideNotebook
from yt.utilities.minimal_representation import MinimalSliceData
from yt.utilities.orientation import Orientation


class YTSlice(YTSelectionContainer2D):
    """
    This is a data object corresponding to a slice through the simulation
    domain.

    This object is typically accessed through the `slice` object that hangs
    off of index objects.  Slice is an orthogonal slice through the
    data, taking all the points at the finest resolution available and then
    indexing them.  It is more appropriately thought of as a slice
    'operator' than an object, however, as its field and coordinate can
    both change.

    Parameters
    ----------
    axis : int or char
        The axis along which to slice.  Can be 0, 1, or 2 for x, y, z.
    coord : float
        The coordinate along the axis at which to slice.  This is in
        "domain" coordinates.
    center : array_like, optional
        The 'center' supplied to fields that use it.  Note that this does
        not have to have `coord` as one value.  optional.
    ds: ~yt.data_objects.static_output.Dataset, optional
        An optional dataset to use rather than self.ds
    field_parameters : dictionary
         A dictionary of field parameters than can be accessed by derived
         fields.
    data_source: optional
        Draw the selection from the provided data source rather than
        all data associated with the data_set

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("RedshiftOutput0005")
    >>> slice = ds.slice(0, 0.25)
    >>> print(slice["gas", "density"])
    """

    _top_node = "/Slices"
    _type_name = "slice"
    _con_args = ("axis", "coord")
    _container_fields = ("px", "py", "pz", "pdx", "pdy", "pdz")

    def __init__(
        self, axis, coord, center=None, ds=None, field_parameters=None, data_source=None
    ):
        validate_axis(ds, axis)
        validate_float(coord)
        # center is an optional parameter
        if center is not None:
            validate_center(center)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        YTSelectionContainer2D.__init__(self, axis, ds, field_parameters, data_source)
        self._set_center(center)
        self.coord = fix_length(coord, ds)

    def _generate_container_field(self, field):
        xax = self.ds.coordinates.x_axis[self.axis]
        yax = self.ds.coordinates.y_axis[self.axis]
        if self._current_chunk is None:
            self.index._identify_base_chunk(self)
        if field == "px":
            return self._current_chunk.fcoords[:, xax]
        elif field == "py":
            return self._current_chunk.fcoords[:, yax]
        elif field == "pz":
            return self._current_chunk.fcoords[:, self.axis]
        elif field == "pdx":
            return self._current_chunk.fwidth[:, xax] * 0.5
        elif field == "pdy":
            return self._current_chunk.fwidth[:, yax] * 0.5
        elif field == "pdz":
            return self._current_chunk.fwidth[:, self.axis] * 0.5
        else:
            raise KeyError(field)

    @property
    def _mrep(self):
        return MinimalSliceData(self)

    def to_pw(self, fields=None, center="center", width=None, origin="center-window"):
        r"""Create a :class:`~yt.visualization.plot_window.PWViewerMPL` from this
        object.

        This is a bare-bones mechanism of creating a plot window from this
        object, which can then be moved around, zoomed, and on and on.  All
        behavior of the plot window is relegated to that routine.
        """
        pw = self._get_pw(fields, center, width, origin, "Slice")
        return pw

    def plot(self, fields=None):
        if hasattr(self._data_source, "left_edge") and hasattr(
            self._data_source, "right_edge"
        ):
            left_edge = self._data_source.left_edge
            right_edge = self._data_source.right_edge
            center = (left_edge + right_edge) / 2.0
            width = right_edge - left_edge
            xax = self.ds.coordinates.x_axis[self.axis]
            yax = self.ds.coordinates.y_axis[self.axis]
            lx, rx = left_edge[xax], right_edge[xax]
            ly, ry = left_edge[yax], right_edge[yax]
            width = (rx - lx), (ry - ly)
        else:
            width = self.ds.domain_width
            center = self.ds.domain_center
        pw = self._get_pw(fields, center, width, "native", "Slice")
        try:
            pw.show()
        except YTNotInsideNotebook:
            pass
        return pw


class YTCuttingPlane(YTSelectionContainer2D):
    """
    This is a data object corresponding to an oblique slice through the
    simulation domain.

    This object is typically accessed through the `cutting` object
    that hangs off of index objects.  A cutting plane is an oblique
    plane through the data, defined by a normal vector and a coordinate.
    It attempts to guess an 'north' vector, which can be overridden, and
    then it pixelizes the appropriate data onto the plane without
    interpolation.

    Parameters
    ----------
    normal : array_like
        The vector that defines the desired plane.  For instance, the
        angular momentum of a sphere.
    center : array_like
        The center of the cutting plane, where the normal vector is anchored.
    north_vector: array_like, optional
        An optional vector to describe the north-facing direction in the resulting
        plane.
    ds: ~yt.data_objects.static_output.Dataset, optional
        An optional dataset to use rather than self.ds
    field_parameters : dictionary
         A dictionary of field parameters than can be accessed by derived
         fields.
    data_source: optional
        Draw the selection from the provided data source rather than
        all data associated with the dataset

    Notes
    -----

    This data object in particular can be somewhat expensive to create.
    It's also important to note that unlike the other 2D data objects, this
    object provides px, py, pz, as some cells may have a height from the
    plane.

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("RedshiftOutput0005")
    >>> cp = ds.cutting([0.1, 0.2, -0.9], [0.5, 0.42, 0.6])
    >>> print(cp["gas", "density"])
    """

    _plane = None
    _top_node = "/CuttingPlanes"
    _key_fields = YTSelectionContainer2D._key_fields + ["pz", "pdz"]
    _type_name = "cutting"
    _con_args = ("normal", "center")
    _tds_attrs = ("_inv_mat",)
    _tds_fields = ("x", "y", "z", "dx")
    _container_fields = ("px", "py", "pz", "pdx", "pdy", "pdz")

    def __init__(
        self,
        normal,
        center,
        north_vector=None,
        ds=None,
        field_parameters=None,
        data_source=None,
    ):
        validate_3d_array(normal)
        validate_center(center)
        if north_vector is not None:
            validate_3d_array(north_vector)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        YTSelectionContainer2D.__init__(self, None, ds, field_parameters, data_source)
        self._set_center(center)
        self.set_field_parameter("center", self.center)
        # Let's set up our plane equation
        # ax + by + cz + d = 0
        self.orienter = Orientation(normal, north_vector=north_vector)
        self._norm_vec = self.orienter.normal_vector
        self._d = -1.0 * np.dot(self._norm_vec, self.center)
        self._x_vec = self.orienter.unit_vectors[0]
        self._y_vec = self.orienter.unit_vectors[1]
        # First we try all three, see which has the best result:
        self._rot_mat = np.array([self._x_vec, self._y_vec, self._norm_vec])
        self._inv_mat = np.linalg.pinv(self._rot_mat)
        self.set_field_parameter("cp_x_vec", self._x_vec)
        self.set_field_parameter("cp_y_vec", self._y_vec)
        self.set_field_parameter("cp_z_vec", self._norm_vec)

    @property
    def normal(self):
        return self._norm_vec

    def _generate_container_field(self, field):
        if self._current_chunk is None:
            self.index._identify_base_chunk(self)
        if field == "px":
            x = self._current_chunk.fcoords[:, 0] - self.center[0]
            y = self._current_chunk.fcoords[:, 1] - self.center[1]
            z = self._current_chunk.fcoords[:, 2] - self.center[2]
            tr = np.zeros(x.size, dtype="float64")
            tr = self.ds.arr(tr, "code_length")
            tr += x * self._x_vec[0]
            tr += y * self._x_vec[1]
            tr += z * self._x_vec[2]
            return tr
        elif field == "py":
            x = self._current_chunk.fcoords[:, 0] - self.center[0]
            y = self._current_chunk.fcoords[:, 1] - self.center[1]
            z = self._current_chunk.fcoords[:, 2] - self.center[2]
            tr = np.zeros(x.size, dtype="float64")
            tr = self.ds.arr(tr, "code_length")
            tr += x * self._y_vec[0]
            tr += y * self._y_vec[1]
            tr += z * self._y_vec[2]
            return tr
        elif field == "pz":
            x = self._current_chunk.fcoords[:, 0] - self.center[0]
            y = self._current_chunk.fcoords[:, 1] - self.center[1]
            z = self._current_chunk.fcoords[:, 2] - self.center[2]
            tr = np.zeros(x.size, dtype="float64")
            tr = self.ds.arr(tr, "code_length")
            tr += x * self._norm_vec[0]
            tr += y * self._norm_vec[1]
            tr += z * self._norm_vec[2]
            return tr
        elif field == "pdx":
            return self._current_chunk.fwidth[:, 0] * 0.5
        elif field == "pdy":
            return self._current_chunk.fwidth[:, 1] * 0.5
        elif field == "pdz":
            return self._current_chunk.fwidth[:, 2] * 0.5
        else:
            raise KeyError(field)

    def to_pw(self, fields=None, center="center", width=None, axes_unit=None):
        r"""Create a :class:`~yt.visualization.plot_window.PWViewerMPL` from this
        object.

        This is a bare-bones mechanism of creating a plot window from this
        object, which can then be moved around, zoomed, and on and on.  All
        behavior of the plot window is relegated to that routine.
        """
        normal = self.normal
        center = self.center
        self.fields = list(iter_fields(fields)) + [
            k for k in self.field_data.keys() if k not in self._key_fields
        ]
        from yt.visualization.fixed_resolution import FixedResolutionBuffer
        from yt.visualization.plot_window import (
            PWViewerMPL,
            get_oblique_window_parameters,
        )

        (bounds, center_rot) = get_oblique_window_parameters(
            normal, center, width, self.ds
        )
        pw = PWViewerMPL(
            self,
            bounds,
            fields=self.fields,
            origin="center-window",
            periodic=False,
            oblique=True,
            frb_generator=FixedResolutionBuffer,
            plot_type="OffAxisSlice",
        )
        if axes_unit is not None:
            pw.set_axes_unit(axes_unit)
        pw._setup_plots()
        return pw

    def to_frb(self, width, resolution, height=None, periodic=False):
        r"""This function returns a FixedResolutionBuffer generated from this
        object.

        An FixedResolutionBuffer is an object that accepts a
        variable-resolution 2D object and transforms it into an NxM bitmap that
        can be plotted, examined or processed.  This is a convenience function
        to return an FRB directly from an existing 2D data object.  Unlike the
        corresponding to_frb function for other YTSelectionContainer2D objects,
        this does not accept a 'center' parameter as it is assumed to be
        centered at the center of the cutting plane.

        Parameters
        ----------
        width : width specifier
            This can either be a floating point value, in the native domain
            units of the simulation, or a tuple of the (value, unit) style.
            This will be the width of the FRB.
        height : height specifier, optional
            This will be the height of the FRB, by default it is equal to width.
        resolution : int or tuple of ints
            The number of pixels on a side of the final FRB.
        periodic : boolean
            This can be true or false, and governs whether the pixelization
            will span the domain boundaries.

        Returns
        -------
        frb : :class:`~yt.visualization.fixed_resolution.FixedResolutionBuffer`
            A fixed resolution buffer, which can be queried for fields.

        Examples
        --------

        >>> v, c = ds.find_max(("gas", "density"))
        >>> sp = ds.sphere(c, (100.0, "au"))
        >>> L = sp.quantities.angular_momentum_vector()
        >>> cutting = ds.cutting(L, c)
        >>> frb = cutting.to_frb((1.0, "pc"), 1024)
        >>> write_image(np.log10(frb["gas", "density"]), "density_1pc.png")
        """
        if is_sequence(width):
            validate_width_tuple(width)
            width = self.ds.quan(width[0], width[1])
        if height is None:
            height = width
        elif is_sequence(height):
            validate_width_tuple(height)
            height = self.ds.quan(height[0], height[1])
        if not is_sequence(resolution):
            resolution = (resolution, resolution)
        from yt.visualization.fixed_resolution import FixedResolutionBuffer

        bounds = (-width / 2.0, width / 2.0, -height / 2.0, height / 2.0)
        frb = FixedResolutionBuffer(self, bounds, resolution, periodic=periodic)
        return frb



from yt.frontends.eagle.api import EagleDataset
from yt.testing import ParticleSelectionComparison, requires_file, requires_module
from yt.utilities.answer_testing.framework import data_dir_load

s28 = "snapshot_028_z000p000/snap_028_z000p000.0.hdf5"


@requires_module("h5py")
@requires_file(s28)
def test_EagleDataset():
    ds = data_dir_load(s28)
    assert isinstance(ds, EagleDataset)
    psc = ParticleSelectionComparison(ds)
    psc.run_defaults()


s399 = "snipshot_399_z000p000/snip_399_z000p000.0.hdf5"


@requires_module("h5py")
@requires_file(s399)
def test_Snipshot():
    ds = data_dir_load(s399)
    assert isinstance(ds, EagleDataset)
    psc = ParticleSelectionComparison(ds)
    psc.run_defaults()



import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.mplot3d import Axes3D  # noqa: F401
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

import yt

# Load the dataset
ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

# Create a sphere object centered on the highest density point in the simulation
# with radius 1 Mpc
sphere = ds.sphere("max", (1.0, "Mpc"))

# Identify the isodensity surface in this sphere with density = 1e-24 g/cm^3
surface = ds.surface(sphere, ("gas", "density"), 1e-24)

# Color this isodensity surface according to the log of the temperature field
colors = yt.apply_colormap(np.log10(surface["gas", "temperature"]), cmap_name="hot")

# Create a 3D matplotlib figure for visualizing the surface
fig = plt.figure()
ax = fig.add_subplot(projection="3d")
p3dc = Poly3DCollection(surface.triangles, linewidth=0.0)

# Set the surface colors in the right scaling [0,1]
p3dc.set_facecolors(colors[0, :, :] / 255.0)
ax.add_collection(p3dc)

# Let's keep the axis ratio fixed in all directions by taking the maximum
# extent in one dimension and make it the bounds in all dimensions
max_extent = (surface.vertices.max(axis=1) - surface.vertices.min(axis=1)).max()
centers = (surface.vertices.max(axis=1) + surface.vertices.min(axis=1)) / 2
bounds = np.zeros([3, 2])
bounds[:, 0] = centers[:] - max_extent / 2
bounds[:, 1] = centers[:] + max_extent / 2
ax.auto_scale_xyz(bounds[0, :], bounds[1, :], bounds[2, :])

# Save the figure
plt.savefig(f"{ds}_Surface.png")



"""
Bit array functions



"""


import numpy as np

cimport cython
cimport numpy as np


cdef inline void ba_set_value(np.uint8_t *buf, np.uint64_t ind,
                              np.uint8_t val) noexcept nogil:
    # This assumes 8 bit buffer.  If value is greater than zero (thus allowing
    # us to use 1-255 as 'True') then we identify first the index in the buffer
    # we are setting.  We do this by truncating the index by bit-shifting to
    # the left three times, essentially dividing it by eight (and taking the
    # floor.)
    # The next step is to turn *on* what we're attempting to turn on, which
    # means taking our index and truncating it to the first 3 bits (which we do
    # with an & operation) and then turning on the correct bit.
    #
    # So if we're asking for index 33 in the bitarray, we would want the 4th
    # uint8 element, then the 2nd bit (index 1).
    #
    # To turn it on, we logical *or* with that.  To turn it off, we logical
    # *and* with the *inverse*, which will allow everything *but* that bit to
    # stay on.
    if val > 0:
        buf[ind >> 3] |= (1 << (ind & 7))
    else:
        buf[ind >> 3] &= ~(1 << (ind & 7))

cdef inline np.uint8_t ba_get_value(np.uint8_t *buf, np.uint64_t ind) noexcept nogil:
    cdef np.uint8_t rv = (buf[ind >> 3] & (1 << (ind & 7)))
    if rv == 0: return 0
    return 1

cdef inline void ba_set_range(np.uint8_t *buf, np.uint64_t start_ind,
                              np.uint64_t stop_ind, np.uint8_t val) nogil:
    # Should this be inclusive of both end points?  I think it should not, to
    # match slicing semantics.
    #
    # We need to figure out the first and last values, and then we just set the
    # ones in-between to 255.
    if stop_ind < start_ind: return
    cdef np.uint64_t i
    cdef np.uint8_t j, bitmask
    cdef np.uint64_t buf_start = start_ind >> 3
    cdef np.uint64_t buf_stop = stop_ind >> 3
    cdef np.uint8_t start_j = start_ind & 7
    cdef np.uint8_t stop_j = stop_ind & 7
    if buf_start == buf_stop:
        for j in range(start_j, stop_j):
            ba_set_value(&buf[buf_start], j, val)
        return
    bitmask = 0
    for j in range(start_j, 8):
        bitmask |= (1 << j)
    if val > 0:
        buf[buf_start] |= bitmask
    else:
        buf[buf_start] &= ~bitmask
    if val > 0:
        bitmask = 255
    else:
        bitmask = 0
    for i in range(buf_start + 1, buf_stop):
        buf[i] = bitmask
    bitmask = 0
    for j in range(0, stop_j):
        bitmask |= (1 << j)
    if val > 0:
        buf[buf_stop] |= bitmask
    else:
        buf[buf_stop] &= ~bitmask


cdef inline np.uint8_t _num_set_bits( np.uint8_t b ):
    # https://stackoverflow.com/questions/30688465/how-to-check-the-number-of-set-bits-in-an-8-bit-unsigned-char
    b = b - ((b >> 1) & 0x55)
    b = (b & 0x33) + ((b >> 2) & 0x33)
    return (((b + (b >> 4)) & 0x0F) * 0x01)

cdef class bitarray:
    cdef np.uint8_t *buf
    cdef np.uint64_t size
    cdef np.uint64_t buf_size # Not exactly the same
    cdef np.uint8_t final_bitmask
    cdef public object ibuf

    cdef void _set_value(self, np.uint64_t ind, np.uint8_t val)
    cdef np.uint8_t _query_value(self, np.uint64_t ind)
    cdef void _set_range(self, np.uint64_t start, np.uint64_t stop, np.uint8_t val)
    cdef np.uint64_t _count(self)
    cdef bitarray _logical_and(self, bitarray other, bitarray result = *)
    cdef bitarray _logical_or(self, bitarray other, bitarray result = *)
    cdef bitarray _logical_xor(self, bitarray other, bitarray result = *)



import glob
import os
from functools import cached_property
from typing import Any, Optional

import numpy as np

from yt.data_objects.static_output import ParticleDataset
from yt.frontends.halo_catalog.data_structures import HaloCatalogFile
from yt.funcs import setdefaultattr
from yt.geometry.particle_geometry_handler import ParticleIndex
from yt.utilities import fortran_utils as fpu
from yt.utilities.cosmology import Cosmology
from yt.utilities.exceptions import YTFieldNotFound

from .definitions import header_dt
from .fields import RockstarFieldInfo


class RockstarBinaryFile(HaloCatalogFile):
    header: dict
    _position_offset: int
    _member_offset: int
    _Npart: "np.ndarray[Any, np.dtype[np.int64]]"
    _ids_halos: list[int]
    _file_size: int

    def __init__(self, ds, io, filename, file_id, range):
        with open(filename, "rb") as f:
            self.header = fpu.read_cattrs(f, header_dt, "=")
            self._position_offset = f.tell()
            pcount = self.header["num_halos"]

            halos = np.fromfile(f, dtype=io._halo_dt, count=pcount)
            self._member_offset = f.tell()
            self._ids_halos = list(halos["particle_identifier"])
            self._Npart = halos["num_p"]

            f.seek(0, os.SEEK_END)
            self._file_size = f.tell()

        expected_end = self._member_offset + 8 * self._Npart.sum()
        if expected_end != self._file_size:
            raise RuntimeError(
                f"File size {self._file_size} does not match expected size {expected_end}."
            )

        super().__init__(ds, io, filename, file_id, range)

    def _read_member(
        self, ihalo: int
    ) -> Optional["np.ndarray[Any, np.dtype[np.int64]]"]:
        if ihalo not in self._ids_halos:
            return None

        ind_halo = self._ids_halos.index(ihalo)

        ipos = self._member_offset + 8 * self._Npart[:ind_halo].sum()

        with open(self.filename, "rb") as f:
            f.seek(ipos, os.SEEK_SET)
            ids = np.fromfile(f, dtype=np.int64, count=self._Npart[ind_halo])
            return ids

    def _read_particle_positions(self, ptype: str, f=None):
        """
        Read all particle positions in this file.
        """

        if f is None:
            close = True
            f = open(self.filename, "rb")
        else:
            close = False

        pcount = self.header["num_halos"]
        pos = np.empty((pcount, 3), dtype="float64")
        f.seek(self._position_offset, os.SEEK_SET)
        halos = np.fromfile(f, dtype=self.io._halo_dt, count=pcount)
        for i, ax in enumerate("xyz"):
            pos[:, i] = halos[f"particle_position_{ax}"].astype("float64")

        if close:
            f.close()

        return pos


class RockstarIndex(ParticleIndex):
    def get_member(self, ihalo: int):
        for df in self.data_files:
            members = df._read_member(ihalo)
            if members is not None:
                return members

        raise RuntimeError(f"Could not find halo {ihalo} in any data file.")


class RockstarDataset(ParticleDataset):
    _index_class = RockstarIndex
    _file_class = RockstarBinaryFile
    _field_info_class = RockstarFieldInfo
    _suffix = ".bin"

    def __init__(
        self,
        filename,
        dataset_type="rockstar_binary",
        units_override=None,
        unit_system="cgs",
        index_order=None,
        index_filename=None,
    ):
        super().__init__(
            filename,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
        )

    def _parse_parameter_file(self):
        with open(self.parameter_filename, "rb") as f:
            hvals = fpu.read_cattrs(f, header_dt)
            hvals.pop("unused")
        self.dimensionality = 3
        self.refine_by = 2
        prefix = ".".join(self.parameter_filename.rsplit(".", 2)[:-2])
        self.filename_template = f"{prefix}.%(num)s{self._suffix}"
        self.file_count = len(glob.glob(prefix + ".*" + self._suffix))

        # Now we can set up things we already know.
        self.cosmological_simulation = 1
        self.current_redshift = (1.0 / hvals["scale"]) - 1.0
        self.hubble_constant = hvals["h0"]
        self.omega_lambda = hvals["Ol"]
        self.omega_matter = hvals["Om"]
        cosmo = Cosmology(
            hubble_constant=self.hubble_constant,
            omega_matter=self.omega_matter,
            omega_lambda=self.omega_lambda,
        )
        self.current_time = cosmo.lookback_time(self.current_redshift, 1e6).in_units(
            "s"
        )
        self._periodicity = (True, True, True)
        self.particle_types = "halos"
        self.particle_types_raw = "halos"

        self.domain_left_edge = np.array([0.0, 0.0, 0.0])
        self.domain_right_edge = np.array([hvals["box_size"]] * 3)

        self.domain_dimensions = np.ones(3, "int32")
        self.parameters.update(hvals)

    def _set_code_unit_attributes(self):
        z = self.current_redshift
        setdefaultattr(self, "length_unit", self.quan(1.0 / (1.0 + z), "Mpc / h"))
        setdefaultattr(self, "mass_unit", self.quan(1.0, "Msun / h"))
        setdefaultattr(self, "velocity_unit", self.quan(1.0, "km / s"))
        setdefaultattr(self, "time_unit", self.length_unit / self.velocity_unit)

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        if not filename.endswith(".bin"):
            return False
        try:
            with open(filename, mode="rb") as f:
                header = fpu.read_cattrs(f, header_dt)
        except OSError:
            return False
        else:
            return header["magic"] == 18077126535843729616

    def halo(self, ptype, particle_identifier):
        return RockstarHaloContainer(
            ptype,
            particle_identifier,
            parent_ds=None,
            halo_ds=self,
        )


class RockstarHaloContainer:
    def __init__(self, ptype, particle_identifier, *, parent_ds, halo_ds):
        if ptype not in halo_ds.particle_types_raw:
            raise RuntimeError(
                f'Possible halo types are {halo_ds.particle_types_raw}, supplied "{ptype}".'
            )

        self.ds = parent_ds
        self.halo_ds = halo_ds
        self.ptype = ptype
        self.particle_identifier = particle_identifier

    def __repr__(self):
        return f"{self.halo_ds}_{self.ptype}_{self.particle_identifier:09d}"

    def __getitem__(self, key):
        if isinstance(key, tuple):
            ptype, field = key
        else:
            ptype = self.ptype
            field = key

        data = {
            "mass": self.mass,
            "position": self.position,
            "velocity": self.velocity,
            "member_ids": self.member_ids,
        }
        if ptype == "halos" and field in data:
            return data[field]

        raise YTFieldNotFound((ptype, field), dataset=self.ds)

    @cached_property
    def ihalo(self):
        halo_id = self.particle_identifier
        halo_ids = list(self.halo_ds.r["halos", "particle_identifier"].astype("i8"))
        ihalo = halo_ids.index(halo_id)

        assert halo_ids[ihalo] == halo_id

        return ihalo

    @property
    def mass(self):
        return self.halo_ds.r["halos", "particle_mass"][self.ihalo]

    @property
    def position(self):
        return self.halo_ds.r["halos", "particle_position"][self.ihalo]

    @property
    def velocity(self):
        return self.halo_ds.r["halos", "particle_velocity"][self.ihalo]

    @property
    def member_ids(self):
        return self.halo_ds.index.get_member(self.particle_identifier)
